import abc

from menpo.transform.base import Transform
from functools import reduce


class ComposableTransform(Transform):
    r"""
    :map:`Transform` subclass that enables native composition, such that
    the behavior of multiple :map:`Transform` s is composed together in a
    natural way.
    """

    @abc.abstractproperty
    def composes_inplace_with(self):
        r"""
        The :map:`Transform` s that this transform composes inplace
        with **natively** (i.e. no :map:`TransformChain` will be produced).

        An attempt to compose inplace against any type that is not an
        instance of this property on this class will result in an `Exception`.

        :type: :map:`Transform` or tuple of :map:`Transform` s
        """

    @property
    def composes_with(self):
        r"""
        The :map:`Transform` s that this transform composes
        with **natively** (i.e. no :map:`TransformChain` will be produced).

        If native composition is not possible, falls back to producing a
        :map:`TransformChain`.

        By default, this is the same list as :attr:`composes_inplace_with`.

        :type: :map:`Transform` or tuple of :map:`Transform` s
        """
        return self.composes_inplace_with

    def compose_before(self, transform):
        r"""
        A :map:`Transform` that represents **this** transform
        composed **before** the given transform::

            c = a.compose_before(b)
            c.apply(p) == b.apply(a.apply(p))

        ``a`` and ``b`` are left unchanged.

        An attempt is made to perform native composition, but will fall back
        to a :map:`TransformChain` as a last resort. See :attr:`composes_with`
        for a description of how the mode of composition is decided.

        Parameters
        ----------
        transform : :map:`Transform`
            Transform to be applied **after** ``self``

        Returns
        --------
        transform : :map:`Transform` or :map:`TransformChain`
            If the composition was native, a single new :map:`Transform` will
            be returned. If not, a :map:`TransformChain` is returned instead.
        """
        if isinstance(transform, self.composes_with):
            return self._compose_before(transform)
        else:
            # best we can do is a TransformChain, let Transform handle that.
            return Transform.compose_before(self, transform)

    def compose_after(self, transform):
        r"""
        A :map:`Transform` that represents **this** transform
        composed **after** the given transform::

            c = a.compose_after(b)
            c.apply(p) == a.apply(b.apply(p))

        ``a`` and ``b`` are left unchanged.

        This corresponds to the usual mathematical formalism for the compose
        operator, `o`.

        An attempt is made to perform native composition, but will fall back
        to a :map:`TransformChain` as a last resort. See :attr:`composes_with`
        for a description of how the mode of composition is decided.

        Parameters
        ----------
        transform : :map:`Transform`
            Transform to be applied **before** ``self``

        Returns
        --------
        transform : :map:`Transform` or :map:`TransformChain`
            If the composition was native, a single new :map:`Transform` will
            be returned. If not, a :map:`TransformChain` is returned instead.
        """
        if isinstance(transform, self.composes_with):
            return self._compose_after(transform)
        else:
            # best we can do is a TransformChain, let Transform handle that.
            return Transform.compose_after(self, transform)

    def compose_before_inplace(self, transform):
        r"""
        Update ``self`` so that it represents **this** transform composed
        **before** the given transform::

            a_orig = a.copy()
            a.compose_before_inplace(b)
            a.apply(p) == b.apply(a_orig.apply(p))

        ``a`` is permanently altered to be the result of the composition.
        ``b`` is left unchanged.

        Parameters
        ----------
        transform : :attr:`composes_inplace_with`
            Transform to be applied **after** ``self``

        Raises
        ------
        ValueError
            If ``transform`` isn't an instance of :attr:`composes_inplace_with`
        """
        if isinstance(transform, self.composes_inplace_with):
            self._compose_before_inplace(transform)
        else:
            raise ValueError(
                "{} can only compose inplace with {} - not "
                "{}".format(type(self), self.composes_inplace_with,
                            type(transform)))

    def compose_after_inplace(self, transform):
        r"""
        Update ``self`` so that it represents **this** transform composed
        **after** the given transform::

            a_orig = a.copy()
            a.compose_after_inplace(b)
            a.apply(p) == a_orig.apply(b.apply(p))

        ``a`` is permanently altered to be the result of the composition. ``b``
        is left unchanged.

        Parameters
        ----------
        transform : :attr:`composes_inplace_with`
            Transform to be applied **before** ``self``

        Raises
        ------
        ValueError
            If ``transform`` isn't an instance of :attr:`composes_inplace_with`
        """
        if isinstance(transform, self.composes_inplace_with):
            self._compose_after_inplace(transform)
        else:
            raise ValueError(
                "{} can only compose inplace with {} - not "
                "{}".format(type(self), self.composes_inplace_with,
                            type(transform)))

    def _compose_before(self, transform):
        r"""
        Naive implementation of composition, ``self.copy()`` and then
        :meth:``compose_before_inplace``. Apply this transform **first**.

        Parameters
        ----------
        transform : :map:`ComposableTransform`
            Transform to be applied **after** ``self``

        Returns
        --------
        transform : :map:`ComposableTransform`
            The resulting transform.
        """
        # naive approach - copy followed by the inplace operation
        self_copy = self.copy()
        self_copy._compose_before_inplace(transform)
        return self_copy

    def _compose_after(self, transform):
        r"""
        Naive implementation of composition, ``self.copy()`` and then
        :meth:``compose_after_inplace``. Apply this transform **second**.

        Parameters
        ----------
        transform : :map:`ComposableTransform`
            Transform to be applied **before** ``self``

        Returns
        --------
        transform : :map:`ComposableTransform`
            The resulting transform.
        """
        # naive approach - copy followed by the inplace operation
        self_copy = self.copy()
        self_copy._compose_after_inplace(transform)
        return self_copy

    @abc.abstractmethod
    def _compose_before_inplace(self, transform):
        r"""
        Specialised inplace composition. This should be overridden to
        provide specific cases of composition as defined in
        :attr:`composes_inplace_with`.

        Parameters
        ----------
        transform : :attr:`composes_inplace_with`
            Transform to be applied **after** ``self``
        """

    @abc.abstractmethod
    def _compose_after_inplace(self, transform):
        r"""
        Specialised inplace composition. This should be overridden to
        provide specific cases of composition as defined in
        :attr:`composes_inplace_with`.

        Parameters
        ----------
        transform : :attr:`composes_inplace_with`
            Transform to be applied **before** ``self``
        """


class VComposable(object):
    r"""
    Mix-in for :map:`Vectorizable` :map:`ComposableTransform` s.

    Use this mix-in with :map:`ComposableTransform` if the
    :map:`ComposableTransform` in question is :map:`Vectorizable` as this adds
    :meth:`from_vector` variants to the :map:`ComposableTransform` interface.
    These can be tuned for performance.
    """

    @abc.abstractmethod
    def compose_after_from_vector_inplace(self, vector):
        r"""
        Specialised inplace composition with a vector. This should be
        overridden to provide specific cases of composition whereby the current
        state of the transform can be derived purely from the provided vector.

        Parameters
        ----------
        vector : ``(n_parameters,)`` ndarray
            Vector to update the transform state with.
        """


class TransformChain(ComposableTransform):
    r"""
    A chain of transforms that can be efficiently applied one after the
    other.

    This class is the natural product of composition. Note that objects may
    know how to compose themselves more efficiently - such objects
    implement the :map:`ComposableTransform` or :map:`VComposable` interfaces.

    Parameters
    ----------
    transforms : `list` of :map:`Transform`
        The list of transforms to be applied. Note that the first transform
        will be applied first - the result of which is fed into the second
        transform and so on until the chain is exhausted.
    """

    def __init__(self, transforms):
        # TODO Should TransformChain copy on input?
        self.transforms = transforms

    def _apply(self, x, **kwargs):
        r"""
        Applies each of the transforms to the array ``x``, in order.

        Parameters
        ----------
        x : ``(n_points, n_dims)`` `ndarray`
            The array to transform.

        Returns
        -------
        transformed : ``(n_points, n_dims_output)`` `ndarray`
            Transformed array having passed through the chain of transforms.
        """
        return reduce(lambda x_i, tr: tr._apply(x_i), self.transforms, x)

    @property
    def composes_inplace_with(self):
        r"""
        The :map:`Transform` s that this transform composes inplace
        with **natively** (i.e. no :map:`TransformChain` will be produced).

        An attempt to compose inplace against any type that is not an
        instance of this property on this class will result in an `Exception`.

        :type: :map:`Transform` or tuple of :map:`Transform` s
        """
        return Transform

    def _compose_before_inplace(self, transform):
        r"""
        Specialised inplace composition. In this case we merely keep a list
        of :map:`Transform` s to apply in order.

        Parameters
        ----------
        transform : :map:`ComposableTransform`
            Transform to be applied **after** ``self``
        """
        self.transforms.append(transform)

    def _compose_after_inplace(self, transform):
        r"""
        Specialised inplace composition. In this case we merely keep a list
        of :map:`Transform`s to apply in order.

        Parameters
        ----------
        transform : :map:`ComposableTransform`
            Transform to be applied **before** ``self``
        """
        self.transforms.insert(0, transform)



import abc
from collections import OrderedDict, MutableMapping

import numpy as np

from menpo.base import Copyable
from menpo.transform.base import Transformable
from menpo.visualize.base import Viewable


class Landmarkable(Copyable):
    r"""
    Abstract interface for object that can have landmarks attached to them.
    Landmarkable objects have a public dictionary of landmarks which are
    managed by a :map:`LandmarkManager`. This means that
    different sets of landmarks can be attached to the same object.
    Landmarks can be N-dimensional and are expected to be some
    subclass of :map:`PointCloud`. These landmarks
    are wrapped inside a :map:`LandmarkGroup` object that performs
    useful tasks like label filtering and viewing.
    """

    __metaclass__ = abc.ABCMeta

    def __init__(self):
        self._landmarks = None

    @abc.abstractproperty
    def n_dims(self):
        """
        The total number of dimensions.

        :type: `int`
        """
        pass

    @property
    def landmarks(self):
        """
        The landmarks object.

        :type: :map:`LandmarkManager`
        """
        if self._landmarks is None:
            self._landmarks = LandmarkManager()
        return self._landmarks

    @property
    def has_landmarks(self):
        """
        Whether the object has landmarks.

        :type: `bool`
        """
        return self._landmarks is not None

    @landmarks.setter
    def landmarks(self, value):
        """
        Landmarks setter.

        Parameters
        ----------
        value : :map:`LandmarkManager`
            The landmarks to set.
        """
        # firstly, make sure the dim is correct. Note that the dim can be None
        lm_n_dims = value.n_dims
        if lm_n_dims is not None and lm_n_dims != self.n_dims:
            raise ValueError(
                "Trying to set {}D landmarks on a "
                "{}D object".format(value.n_dims, self.n_dims))
        self._landmarks = value.copy()

    @property
    def n_landmark_groups(self):
        r"""
        The number of landmark groups on this object.

        :type: `int`
        """
        return self.landmarks.n_groups


class LandmarkManager(MutableMapping, Transformable):
    """Store for :map:`LandmarkGroup` instances associated with an object

    Every :map:`Landmarkable` instance has an instance of this class available
    at the ``.landmarks`` property.  It is through this class that all access
    to landmarks attached to instances is handled. In general the
    :map:`LandmarkManager` provides a dictionary-like interface for storing
    landmarks. :map:`LandmarkGroup` instances are stored under string keys -
    these keys are refereed to as the **group name**. A special case is
    where there is a single unambiguous :map:`LandmarkGroup` attached to a
    :map:`LandmarkManager` - in this case ``None`` can be used as a key to
    access the sole group.


    Note that all landmarks stored on a :map:`Landmarkable` in it's attached
    :map:`LandmarkManager` are automatically transformed and copied with their
    parent object.
    """
    def __init__(self):
        super(LandmarkManager, self).__init__()
        self._landmark_groups = {}

    @property
    def n_dims(self):
        """
        The total number of dimensions.

        :type: `int`
        """
        if self.n_groups != 0:
            # Python version independent way of getting the first value
            for v in self._landmark_groups.values():
                return v.n_dims
        else:
            return None

    def copy(self):
        r"""
        Generate an efficient copy of this :map:`LandmarkManager`.

        Returns
        -------
        ``type(self)``
            A copy of this object

        """
        # do a normal copy. The dict will be shallow copied - rectify that here
        new = Copyable.copy(self)
        for k, v in new._landmark_groups.items():
            new._landmark_groups[k] = v.copy()
        return new

    def __iter__(self):
        """
        Iterate over the internal landmark group dictionary
        """
        return iter(self._landmark_groups)

    def __setitem__(self, group, value):
        """
        Sets a new landmark group for the given label. This can be set using
        an existing landmark group, or using a PointCloud. Existing landmark
        groups will have their target reset. If a PointCloud is provided then
        all landmarks belong to a single label `all`.

        Parameters
        ----------
        group : `string`
            Label of new group.

        value : :map:`LandmarkGroup` or :map:`PointCloud`
            The new landmark group to set.

        Raises
        ------
        DimensionalityError
            If the landmarks and the shape are not of the same dimensionality.
        """
        from menpo.shape import PointCloud
        # firstly, make sure the dim is correct
        n_dims = self.n_dims
        if n_dims is not None and value.n_dims != n_dims:
            raise ValueError(
                "Trying to set {}D landmarks on a "
                "{}D LandmarkManager".format(value.n_dims, self.n_dims))
        if isinstance(value, PointCloud):
            # Copy the PointCloud so that we take ownership of the memory
            lmark_group = LandmarkGroup(
                value,
                OrderedDict([('all', np.ones(value.n_points, dtype=np.bool))]))
        elif isinstance(value, LandmarkGroup):
            # Copy the landmark group so that we now own it
            lmark_group = value.copy()
            # check the target is set correctly
            lmark_group._group_label = group
        else:
            raise ValueError('Valid types are PointCloud or LandmarkGroup')

        self._landmark_groups[group] = lmark_group

    def __getitem__(self, group=None):
        """
        Returns the group for the provided label.

        Parameters
        ---------
        group : `string`, optional
            The label of the group. If None is provided, and if there is only
            one group, the unambiguous group will be returned.
        Returns
        -------
        lmark_group : :map:`LandmarkGroup`
            The matching landmark group.
        """
        if group is None:
            if self.n_groups == 1:
                group = self.group_labels[0]
            else:
                raise ValueError("Cannot use None as a key as there are {} "
                                 "landmark groups".format(self.n_groups))
        return self._landmark_groups[group]

    def __delitem__(self, group):
        """
        Delete the group for the provided label.

        Parameters
        ---------
        group : `string`
            The label of the group.
        """
        del self._landmark_groups[group]

    def __len__(self):
        return len(self._landmark_groups)

    @property
    def n_groups(self):
        """
        Total number of labels.

        :type: `int`
        """
        return len(self._landmark_groups)

    @property
    def has_landmarks(self):
        """
        Whether the object has landmarks or not

        :type: `int`
        """
        return self.n_groups != 0

    @property
    def group_labels(self):
        """
        All the labels for the landmark set.

        :type: `list` of `str`
        """
        return self._landmark_groups.keys()

    def _transform_inplace(self, transform):
        for group in self._landmark_groups.values():
            group.lms._transform_inplace(transform)
        return self

    def view_widget(self, popup=False, browser_style='buttons',
                    figure_size=(10, 8)):
        r"""
        Visualizes the landmark manager object using the
        :map:`visualize_landmarks` widget.

        Parameters
        ----------
        popup : `bool`, optional
            If ``True``, the widget will appear as a popup window.
        browser_style : {``buttons``, ``slider``}, optional
            It defines whether the selector of the landmark managers will have
            the form of plus/minus buttons or a slider.
        figure_size : (`int`, `int`), optional
            The initial size of the rendered figure.
        """
        from menpo.visualize import visualize_landmarks
        visualize_landmarks(self, figure_size=figure_size, popup=popup,
                            browser_style=browser_style)

    def __str__(self):
        out_string = '{}: n_groups: {}'.format(type(self).__name__,
                                               self.n_groups)
        if self.has_landmarks:
            for label in self:
                out_string += '\n'
                out_string += '({}): {}'.format(label, self[label].__str__())

        return out_string


class LandmarkGroup(MutableMapping, Copyable, Viewable):
    r"""
    An immutable object that holds a :map:`PointCloud` (or a subclass) and
    stores labels for each point. These labels are defined via masks on the
    :map:`PointCloud`. For this reason, the :map:`PointCloud` is considered to
    be immutable.

    The labels to masks must be within an `OrderedDict` so that semantic
    ordering can be maintained.

    Parameters
    ----------
    pointcloud : :map:`PointCloud`
        The pointcloud representing the landmarks.
    labels_to_masks : `ordereddict` {`str` -> `bool ndarray`}
        For each label, the mask that specifies the indices in to the
        pointcloud that belong to the label.
    copy : `bool`, optional
        If ``True``, a copy of the :map:`PointCloud` is stored on the group.

    Raises
    ------
    ValueError
        If `dict` passed instead of `OrderedDict`
    ValueError
        If no set of label masks is passed.
    ValueError
        If any of the label masks differs in size to the pointcloud.
    ValueError
        If there exists any point in the pointcloud that is not covered
        by a label.
    """
    def __init__(self, pointcloud, labels_to_masks, copy=True):
        super(LandmarkGroup, self).__init__()

        if not labels_to_masks:
            raise ValueError('Landmark groups are designed for their internal '
                             'state, other than owernship, to be immutable. '
                             'Empty label sets are not permitted.')
        if np.vstack(labels_to_masks.values()).shape[1] != pointcloud.n_points:
            raise ValueError('Each mask must have the same number of points '
                             'as the landmark pointcloud.')
        if type(labels_to_masks) is dict:
            raise ValueError('Must provide an OrderedDict to maintain the '
                             'semantic meaning of the labels.')

        # Another sanity check
        self._labels_to_masks = labels_to_masks
        self._verify_all_labels_masked()

        if copy:
            self._pointcloud = pointcloud.copy()
            self._labels_to_masks = OrderedDict([(l, m.copy()) for l, m in
                                                 labels_to_masks.items()])
        else:
            self._pointcloud = pointcloud
            self._labels_to_masks = labels_to_masks

    def copy(self):
        r"""
        Generate an efficient copy of this :map:`LandmarkGroup`.

        Returns
        -------
        ``type(self)``
            A copy of this object
        """
        new = Copyable.copy(self)
        for k, v in new._labels_to_masks.items():
            new._labels_to_masks[k] = v.copy()
        return new

    def __iter__(self):
        """
        Iterate over the internal label dictionary
        """
        return iter(self._labels_to_masks)

    def __setitem__(self, label, indices):
        """
        Add a new label to the landmark group by adding a new set of indices

        Parameters
        ----------
        label : `string`
            Label of landmark.

        indices : ``(K,)`` `ndarray`
            Array of indices in to the pointcloud. Each index implies
            membership to the label.
        """
        mask = np.zeros(self._pointcloud.n_points, dtype=np.bool)
        mask[indices] = True
        self._labels_to_masks[label] = mask

    def __getitem__(self, label=None):
        """
        Returns the PointCloud that contains this label represents on the group.
        This will be a subset of the total landmark group PointCloud.

        Parameters
        ----------
        label : `string`
            Label to filter on.

        Returns
        -------
        pcloud : :map:`PointCloud`
            The PointCloud that this label represents. Will be a subset of the
            entire group's landmarks.
        """
        if label is None:
            return self.lms.copy()
        return self._pointcloud.from_mask(self._labels_to_masks[label])

    def __delitem__(self, label):
        """
        Delete the semantic labelling for the provided label.

         .. note::

             You cannot delete a semantic label and leave the landmark group
             partially unlabelled. Landmark groups must contain labels for
             every point.

        Parameters
        ---------
        label : `string`
            The label to remove.

        Raises
        ------
        ValueError
            If deleting the label would leave some points unlabelled
        """
        # Pop the value off, which is akin to deleting it (removes it from the
        # underlying dict). However, we keep it around so we can check if
        # removing it causes an unlabelled point
        value_to_delete = self._labels_to_masks.pop(label)

        try:
            self._verify_all_labels_masked()
        except ValueError as e:
            # Catch the error, restore the value and re-raise the exception!
            self._labels_to_masks[label] = value_to_delete
            raise e

    def __len__(self):
        return len(self._labels_to_masks)

    @property
    def labels(self):
        """
        The list of labels that belong to this group.

        :type: `list` of `str`
        """
        return self._labels_to_masks.keys()

    @property
    def n_labels(self):
        """
        Number of labels in the group.

        :type: `int`
        """
        return len(self.labels)

    @property
    def lms(self):
        """
        The pointcloud representing all the landmarks in the group.

        :type: :map:`PointCloud`
        """
        return self._pointcloud

    @property
    def n_landmarks(self):
        """
        The total number of landmarks in the group.

        :type: `int`
        """
        return self._pointcloud.n_points

    @property
    def n_dims(self):
        """
        The dimensionality of these landmarks.

        :type: `int`
        """
        return self._pointcloud.n_dims

    def with_labels(self, labels=None):
        """A new landmark group that contains only the certain labels

        Parameters
        ----------
        labels : `str` or `list` of `str`, optional
            Labels that should be kept in the returned landmark group. If
            ``None`` is passed, and if there is only one label on this group,
            the label will be substituted automatically.

        Returns
        -------
        landmark_group : :map:`LandmarkGroup`
            A new landmark group with the same group label but containing only
            the given label.
        """
        # make it easier by allowing None when there is only one label
        if labels is None:
            if self.n_labels == 1:
                labels = self.labels
            else:
                raise ValueError("Cannot use None as there are "
                                 "{} labels".format(self.n_labels))
        # Make it easier to use by accepting a single string as well as a list
        if isinstance(labels, str):
            labels = [labels]
        return self._new_group_with_only_labels(labels)

    def without_labels(self, labels):
        """A new landmark group that excludes certain labels
        label.

        Parameters
        ----------
        labels : `str` or `list` of `str`
            Labels that should be excluded in the returned landmark group.

        Returns
        -------
        landmark_group : :map:`LandmarkGroup`
            A new landmark group with the same group label but containing all
            labels except the given label.
        """
        # Make it easier to use by accepting a single string as well as a list
        if isinstance(labels, str):
            labels = [labels]
        labels_to_keep = list(set(self.labels).difference(labels))
        return self._new_group_with_only_labels(labels_to_keep)

    def _verify_all_labels_masked(self):
        """
        Verify that every point in the pointcloud is associated with a label.
        If any one point is not covered by a label, then raise a
        ``ValueError``.
        """
        unlabelled_points = np.sum(self._labels_to_masks.values(), axis=0) == 0
        if np.any(unlabelled_points):
            nonzero = np.nonzero(unlabelled_points)
            raise ValueError(
                'Every point in the landmark pointcloud must be labelled. '
                'Points {0} were unlabelled.'.format(nonzero))

    def _new_group_with_only_labels(self, labels):
        """
        Deal with changing indices when you add and remove points. In this case
        we only deal with building a new dataset that keeps masks.

        Parameters
        ----------
        labels : list of `string`
            List of strings of the labels to keep

        Returns
        -------
        lmark_group : :map:`LandmarkGroup`
            The new landmark group with only the requested labels.
        """
        set_difference = set(labels).difference(self.labels)
        if len(set_difference) > 0:
            raise ValueError('Labels {0} do not exist in the landmark '
                             'group. Available labels are: {1}'.format(
                list(set_difference), self.labels))

        masks_to_keep = [self._labels_to_masks[l] for l in labels
                         if l in self._labels_to_masks]
        overlap = np.sum(masks_to_keep, axis=0) > 0
        masks_to_keep = [l[overlap] for l in masks_to_keep]

        return LandmarkGroup(self._pointcloud.from_mask(overlap),
                             OrderedDict(zip(labels, masks_to_keep)))

    def tojson(self):
        r"""
        Convert this `LandmarkGroup` to a dictionary JSON representation.

        Returns
        -------
        json : ``dict``
            Dictionary conforming to the LJSON v2 specification.
        """
        labels = [{'mask': mask.nonzero()[0].tolist(),
                   'label': label}
                  for label, mask in self._labels_to_masks.items()]

        return {'landmarks': self.lms.tojson(),
                'labels': labels}

    def _view_2d(self, with_labels=None, without_labels=None, group='group',
                 figure_id=None, new_figure=False, image_view=True,
                 render_lines=True, line_colour=None, line_style='-',
                 line_width=1, render_markers=True, marker_style='o',
                 marker_size=20, marker_face_colour=None,
                 marker_edge_colour=None, marker_edge_width=1.,
                 render_numbering=False, numbers_horizontal_align='center',
                 numbers_vertical_align='bottom',
                 numbers_font_name='sans-serif', numbers_font_size=10,
                 numbers_font_style='normal', numbers_font_weight='normal',
                 numbers_font_colour='k', render_legend=True, legend_title='',
                 legend_font_name='sans-serif', legend_font_style='normal',
                 legend_font_size=10, legend_font_weight='normal',
                 legend_marker_scale=None, legend_location=2,
                 legend_bbox_to_anchor=(1.05, 1.), legend_border_axes_pad=None,
                 legend_n_columns=1, legend_horizontal_spacing=None,
                 legend_vertical_spacing=None, legend_border=True,
                 legend_border_padding=None, legend_shadow=False,
                 legend_rounded_corners=False, render_axes=True,
                 axes_font_name='sans-serif', axes_font_size=10,
                 axes_font_style='normal', axes_font_weight='normal',
                 axes_x_limits=None, axes_y_limits=None, figure_size=(10, 8)):
        """
        Visualize the landmark group.

        Parameters
        ----------
        with_labels : ``None`` or `str` or `list` of `str`, optional
            If not ``None``, only show the given label(s). Should **not** be
            used with the ``without_labels`` kwarg.
        without_labels : ``None`` or `str` or `list` of `str`, optional
            If not ``None``, show all except the given label(s). Should **not**
            be used with the ``with_labels`` kwarg.
        group : `str` or `None`, optional
            The landmark group to be visualized. If ``None`` and there are more
            than one landmark groups, an error is raised.
        figure_id : `object`, optional
            The id of the figure to be used.
        new_figure : `bool`, optional
            If ``True``, a new figure is created.
        image_view : `bool`, optional
            If ``True``, the x and y axes are flipped.
        render_lines : `bool`, optional
            If ``True``, the edges will be rendered.
        line_colour : {``r``, ``g``, ``b``, ``c``, ``m``, ``k``, ``w``} or
                      ``(3, )`` `ndarray` or ``None``, optional
            The colour of the lines. If ``None``, a different colour will be
            automatically selected for each label.
        line_style : {``-``, ``--``, ``-.``, ``:``}, optional
            The style of the lines.
        line_width : `float`, optional
            The width of the lines.
        render_markers : `bool`, optional
            If ``True``, the markers will be rendered.
        marker_style : {``.``, ``,``, ``o``, ``v``, ``^``, ``<``, ``>``, ``+``,
                        ``x``, ``D``, ``d``, ``s``, ``p``, ``*``, ``h``, ``H``,
                        ``1``, ``2``, ``3``, ``4``, ``8``}, optional
            The style of the markers.
        marker_size : `int`, optional
            The size of the markers in points^2.
        marker_face_colour : {``r``, ``g``, ``b``, ``c``, ``m``, ``k``, ``w``}
                             or ``(3, )`` `ndarray`, optional
            The face (filling) colour of the markers.
        marker_edge_colour : {``r``, ``g``, ``b``, ``c``, ``m``, ``k``, ``w``}
                             or ``(3, )`` `ndarray`, optional
            The edge colour of the markers.
        marker_edge_width : `float`, optional
            The width of the markers' edge.
        render_numbering : `bool`, optional
            If ``True``, the landmarks will be numbered.
        numbers_horizontal_align : {``center``, ``right``, ``left``}, optional
            The horizontal alignment of the numbers' texts.
        numbers_vertical_align : {``center``, ``top``, ``bottom``,
                                  ``baseline``}, optional
            The vertical alignment of the numbers' texts.
        numbers_font_name : {``serif``, ``sans-serif``, ``cursive``,
                             ``fantasy``, ``monospace``}, optional
            The font of the numbers.
        numbers_font_size : `int`, optional
            The font size of the numbers.
        numbers_font_style : {``normal``, ``italic``, ``oblique``}, optional
            The font style of the numbers.
        numbers_font_weight : {``ultralight``, ``light``, ``normal``,
                               ``regular``, ``book``, ``medium``, ``roman``,
                               ``semibold``, ``demibold``, ``demi``, ``bold``,
                               ``heavy``, ``extra bold``, ``black``}, optional
            The font weight of the numbers.
        numbers_font_colour : {``r``, ``g``, ``b``, ``c``, ``m``, ``k``, ``w``}
                              or ``(3, )`` `ndarray`, optional
            The font colour of the numbers.
        render_legend : `bool`, optional
            If ``True``, the legend will be rendered.
        legend_title : `str`, optional
            The title of the legend.
        legend_font_name : {``serif``, ``sans-serif``, ``cursive``,
                            ``fantasy``, ``monospace``}, optional
            The font of the legend.
        legend_font_style : {``normal``, ``italic``, ``oblique``}, optional
            The font style of the legend.
        legend_font_size : `int`, optional
            The font size of the legend.
        legend_font_weight : {``ultralight``, ``light``, ``normal``,
                              ``regular``, ``book``, ``medium``, ``roman``,
                              ``semibold``, ``demibold``, ``demi``, ``bold``,
                              ``heavy``, ``extra bold``, ``black``}, optional
            The font weight of the legend.
        legend_marker_scale : `float`, optional
            The relative size of the legend markers with respect to the original
        legend_location : `int`, optional
            The location of the legend. The predefined values are:

            =============== ===
            'best'          0
            'upper right'   1
            'upper left'    2
            'lower left'    3
            'lower right'   4
            'right'         5
            'center left'   6
            'center right'  7
            'lower center'  8
            'upper center'  9
            'center'        10
            =============== ===

        legend_bbox_to_anchor : (`float`, `float`), optional
            The bbox that the legend will be anchored.
        legend_border_axes_pad : `float`, optional
            The pad between the axes and legend border.
        legend_n_columns : `int`, optional
            The number of the legend's columns.
        legend_horizontal_spacing : `float`, optional
            The spacing between the columns.
        legend_vertical_spacing : `float`, optional
            The vertical space between the legend entries.
        legend_border : `bool`, optional
            If ``True``, a frame will be drawn around the legend.
        legend_border_padding : `float`, optional
            The fractional whitespace inside the legend border.
        legend_shadow : `bool`, optional
            If ``True``, a shadow will be drawn behind legend.
        legend_rounded_corners : `bool`, optional
            If ``True``, the frame's corners will be rounded (fancybox).
        render_axes : `bool`, optional
            If ``True``, the axes will be rendered.
        axes_font_name : {``serif``, ``sans-serif``, ``cursive``, ``fantasy``,
                          ``monospace``}, optional
            The font of the axes.
        axes_font_size : `int`, optional
            The font size of the axes.
        axes_font_style : {``normal``, ``italic``, ``oblique``}, optional
            The font style of the axes.
        axes_font_weight : {``ultralight``, ``light``, ``normal``, ``regular``,
                            ``book``, ``medium``, ``roman``, ``semibold``,
                            ``demibold``, ``demi``, ``bold``, ``heavy``,
                            ``extra bold``, ``black``}, optional
            The font weight of the axes.
        axes_x_limits : (`float`, `float`) or `None`, optional
            The limits of the x axis.
        axes_y_limits : (`float`, `float`) or `None`, optional
            The limits of the y axis.
        figure_size : (`float`, `float`) or `None`, optional
            The size of the figure in inches.

        Raises
        ------
        ValueError
            If both ``with_labels`` and ``without_labels`` are passed.
        """
        from menpo.visualize import LandmarkViewer2d
        if with_labels is not None and without_labels is not None:
            raise ValueError('You may only pass one of `with_labels` or '
                             '`without_labels`.')
        elif with_labels is not None:
            lmark_group = self.with_labels(with_labels)
        elif without_labels is not None:
            lmark_group = self.without_labels(without_labels)
        else:
            lmark_group = self  # Fall through
        landmark_viewer = LandmarkViewer2d(figure_id, new_figure,
                                           group, lmark_group._pointcloud,
                                           lmark_group._labels_to_masks)
        return landmark_viewer.render(
            image_view=image_view, render_lines=render_lines,
            line_colour=line_colour, line_style=line_style,
            line_width=line_width, render_markers=render_markers,
            marker_style=marker_style, marker_size=marker_size,
            marker_face_colour=marker_face_colour,
            marker_edge_colour=marker_edge_colour,
            marker_edge_width=marker_edge_width,
            render_numbering=render_numbering,
            numbers_horizontal_align=numbers_horizontal_align,
            numbers_vertical_align=numbers_vertical_align,
            numbers_font_name=numbers_font_name,
            numbers_font_size=numbers_font_size,
            numbers_font_style=numbers_font_style,
            numbers_font_weight=numbers_font_weight,
            numbers_font_colour=numbers_font_colour,
            render_legend=render_legend, legend_title=legend_title,
            legend_font_name=legend_font_name,
            legend_font_style=legend_font_style,
            legend_font_size=legend_font_size,
            legend_font_weight=legend_font_weight,
            legend_marker_scale=legend_marker_scale,
            legend_location=legend_location,
            legend_bbox_to_anchor=legend_bbox_to_anchor,
            legend_border_axes_pad=legend_border_axes_pad,
            legend_n_columns=legend_n_columns,
            legend_horizontal_spacing=legend_horizontal_spacing,
            legend_vertical_spacing=legend_vertical_spacing,
            legend_border=legend_border,
            legend_border_padding=legend_border_padding,
            legend_shadow=legend_shadow,
            legend_rounded_corners=legend_rounded_corners,
            render_axes=render_axes, axes_font_name=axes_font_name,
            axes_font_size=axes_font_size, axes_font_style=axes_font_style,
            axes_font_weight=axes_font_weight, axes_x_limits=axes_x_limits,
            axes_y_limits=axes_y_limits, figure_size=figure_size)

    def _view_3d(self, figure_id=None, new_figure=False, **kwargs):
        try:
            from menpo3d.visualize import LandmarkViewer3d
            return LandmarkViewer3d(figure_id, new_figure,
                                    self._pointcloud, self).render(**kwargs)
        except ImportError:
            from menpo.visualize import Menpo3dErrorMessage
            raise ImportError(Menpo3dErrorMessage)

    def view_widget(self, popup=False, browser_style='buttons',
                    figure_size=(10, 8)):
        r"""
        Visualizes the landmark group object using the
        :map:`visualize_landmarkgroups` widget.

        Parameters
        ----------
        popup : `bool`, optional
            If ``True``, the widget will appear as a popup window.
        browser_style : {``buttons``, ``slider``}, optional
            It defines whether the selector of the landmark managers will have
            the form of plus/minus buttons or a slider.
        figure_size : (`int`, `int`), optional
            The initial size of the rendered figure.
        """
        from menpo.visualize import visualize_landmarkgroups
        visualize_landmarkgroups(self, figure_size=figure_size, popup=popup,
                                 browser_style=browser_style)

    def __str__(self):
        return '{}: n_labels: {}, n_points: {}'.format(
            type(self).__name__, self.n_labels, self.n_landmarks)



from .base import (import_image, import_images,
                   import_builtin_asset,
                   import_landmark_file, import_landmark_files,
                   data_path_to, data_dir_path, ls_builtin_assets,
                   image_paths, landmark_file_paths,
                   import_pickle, import_pickles)



from collections import OrderedDict
import numpy as np
from menpo.landmark.base import LandmarkGroup
from menpo.landmark.exceptions import LabellingError


def _connectivity_from_array(array, close_loop=False):
    r"""
    Build the connectivity over a given array. For example, given ::

        array = [(0, 3, 1, 2 )]

    Generate the connectivity of ::

        [(0, 3), (3, 1), (1, 2)]

    If ``close_loop`` is true, add an extra connection from the last point to
    the first.
    """
    conn = zip(array, array[1:])
    if close_loop:
        conn.append((array[-1], array[0]))
    return np.asarray(conn)


def _connectivity_from_range(range_tuple, close_loop=False):
    r"""
    Build the connectivity over a range. For example, given ::

        range_array = np.arange(3)

    Generate the connectivity of ::

        [(0, 1), (1, 2), (2, 3)]

    If ``close_loop`` is true, add an extra connection from the last point to
    the first.
    """
    return _connectivity_from_array(
        np.arange(*range_tuple), close_loop=close_loop)


def _mask_from_range(range_tuple, n_points):
    r"""
    Generate a mask over the range. The mask will be true inside the range.
    """
    mask = np.zeros(n_points, dtype=np.bool)
    range_slice = slice(*range_tuple)
    mask[range_slice] = True
    return mask


def _build_labelling_error_msg(group, n_expected_points,
                               n_actual_points):
    return '{} mark-up expects exactly {} ' \
           'points. However, the given landmark group ' \
           'has {} points'.format(group, n_expected_points,
                                  n_actual_points)


def _validate_input(landmark_group, n_expected_points, group):
    r"""
    Ensure that the input matches the number of expected points.

    Parameters
    ----------
    landmark_group : :map:`LandmarkGroup`
        Landmark group to validate
    n_expected_points : `int`
        Number of expected points
    group : `str`
        Group label for error message

    Raises
    ------
    LabellingError
        If the number of expected points doesn't match the number of given
        points
    """
    n_points = landmark_group.lms.n_points
    if n_points != n_expected_points:
        raise LabellingError(_build_labelling_error_msg(group,
                                                        n_expected_points,
                                                        n_points))


def _relabel_group_from_dict(pointcloud, labels_to_ranges):
    """
    Label the given pointcloud according to the given ordered dictionary
    of labels to ranges. This assumes that you can semantically label the group
    by using ranges in to the existing points e.g ::

        labels_to_ranges = {'jaw': (0, 17, False)}

    The third element of the range tuple is whether the range is a closed loop
    or not. For example, for an eye landmark this would be ``True``, as you
    do want to create a closed loop for an eye.

    Parameters
    ----------
    pointcloud : :map:`PointCloud`
        The pointcloud to apply semantic labels to.
    labels_to_ranges : `OrderedDict`
        Ordered dictionary of string labels to range tuples.

    Returns
    -------
    landmark_group: :map:`LandmarkGroup`
        New landmark group

    Raises
    ------
    :class:`menpo.landmark.exceptions.LabellingError`
        If the given pointcloud contains less than ``n_expected_points``
        points.
    """
    from menpo.shape import PointUndirectedGraph

    n_points = pointcloud.n_points
    masks = OrderedDict()
    adjacency_lists = []
    for label, tup in labels_to_ranges.items():
        range_tuple = tup[:-1]
        close_loop = tup[-1]
        adjacency_lists.append(_connectivity_from_range(
            range_tuple, close_loop=close_loop))
        masks[label] = _mask_from_range(range_tuple, n_points)
    adjacency_array = np.vstack(adjacency_lists)

    new_landmark_group = LandmarkGroup(
        PointUndirectedGraph(pointcloud.points, adjacency_array), masks)

    return new_landmark_group


def ibug_face_68(landmark_group):
    """
    Apply the ibug's "standard" 68 point semantic labels (based on the
    original semantic labels of multiPIE) to the landmark group.

    The group label will be ``ibug_face_68``.

    The semantic labels applied are as follows:

      - jaw
      - left_eyebrow
      - right_eyebrow
      - nose
      - left_eye
      - right_eye
      - mouth

    Parameters
    ----------
    landmark_group : :map:`LandmarkGroup`
        The landmark group to apply semantic labels to.

    Returns
    -------
    group : `str`
        The group label: ``ibug_face_68``
    landmark_group : :map:`LandmarkGroup`
        New landmark group.

    Raises
    ------
    error : :map:`LabellingError`
        If the given landmark group contains less than 68 points

    References
    ----------
    .. [1] http://www.multipie.org/
    """
    from menpo.shape import PointUndirectedGraph

    group = 'ibug_face_68'
    n_points = 68
    _validate_input(landmark_group, 68, group)

    jaw_indices = np.arange(0, 17)
    lbrow_indices = np.arange(17, 22)
    rbrow_indices = np.arange(22, 27)
    upper_nose_indices = np.arange(27, 31)
    lower_nose_indices = np.arange(31, 36)
    leye_indices = np.arange(36, 42)
    reye_indices = np.arange(42, 48)
    outer_mouth_indices = np.arange(48, 60)
    inner_mouth_indices = np.arange(60, 68)

    jaw_connectivity = _connectivity_from_array(jaw_indices)
    lbrow_connectivity = _connectivity_from_array(lbrow_indices)
    rbrow_connectivity = _connectivity_from_array(rbrow_indices)
    nose_connectivity = np.vstack([
        _connectivity_from_array(upper_nose_indices),
        _connectivity_from_array(lower_nose_indices)])
    leye_connectivity = _connectivity_from_array(leye_indices, close_loop=True)
    reye_connectivity = _connectivity_from_array(reye_indices, close_loop=True)
    mouth_connectivity = np.vstack([
        _connectivity_from_array(outer_mouth_indices, close_loop=True),
        _connectivity_from_array(inner_mouth_indices, close_loop=True)])

    total_conn = np.vstack([
        jaw_connectivity, lbrow_connectivity, rbrow_connectivity,
        nose_connectivity, leye_connectivity, reye_connectivity,
        mouth_connectivity
    ])

    new_landmark_group = LandmarkGroup(
        PointUndirectedGraph(landmark_group.lms.points, total_conn),
        OrderedDict([('all', np.ones(n_points, dtype=np.bool))]))

    new_landmark_group['jaw'] = jaw_indices
    new_landmark_group['left_eyebrow'] = lbrow_indices
    new_landmark_group['right_eyebrow'] = rbrow_indices
    new_landmark_group['nose'] = np.hstack((upper_nose_indices,
                                            lower_nose_indices))
    new_landmark_group['left_eye'] = leye_indices
    new_landmark_group['right_eye'] = reye_indices
    new_landmark_group['mouth'] = np.hstack((outer_mouth_indices,
                                             inner_mouth_indices))

    del new_landmark_group['all']  # Remove pointless all group

    return group, new_landmark_group


def ibug_face_66(landmark_group):
    """
    Apply the ibug's "standard" 66 point semantic labels (based on the
    original semantic labels of multiPIE but ignoring the 2 points
    describing the inner mouth corners) to the landmark group.

    The group label will be ``ibug_face_66``.

    The semantic labels applied are as follows:

      - jaw
      - left_eyebrow
      - right_eyebrow
      - nose
      - left_eye
      - right_eye
      - mouth

    Parameters
    ----------
    landmark_group : :map:`LandmarkGroup`
        The landmark group to apply semantic labels to.

    Returns
    -------
    group : `str`
        The group label: ``ibug_face_66``
    landmark_group : :map:`LandmarkGroup`
        New landmark group.

    Raises
    ------
    error : :map:`LabellingError`
        If the given landmark group contains less than 68 points

    References
    ----------
    .. [1] http://www.multipie.org/
    """
    from menpo.shape import PointUndirectedGraph

    group = 'ibug_face_66'
    n_points = 66
    _validate_input(landmark_group, 68, group)

    jaw_indices = np.arange(0, 17)
    lbrow_indices = np.arange(17, 22)
    rbrow_indices = np.arange(22, 27)
    upper_nose_indices = np.arange(27, 31)
    lower_nose_indices = np.arange(31, 36)
    leye_indices = np.arange(36, 42)
    reye_indices = np.arange(42, 48)
    outer_mouth_indices = np.arange(48, 60)
    inner_mouth_indices = np.hstack((48, np.arange(60, 63),
                                     54, np.arange(63, 66)))

    jaw_connectivity = _connectivity_from_array(jaw_indices)
    lbrow_connectivity = _connectivity_from_array(lbrow_indices)
    rbrow_connectivity = _connectivity_from_array(rbrow_indices)
    nose_connectivity = np.vstack([
        _connectivity_from_array(upper_nose_indices),
        _connectivity_from_array(lower_nose_indices)])
    leye_connectivity = _connectivity_from_array(leye_indices, close_loop=True)
    reye_connectivity = _connectivity_from_array(reye_indices, close_loop=True)
    mouth_connectivity = np.vstack([
        _connectivity_from_array(outer_mouth_indices, close_loop=True),
        _connectivity_from_array(inner_mouth_indices, close_loop=True)])

    total_conn = np.vstack([
        jaw_connectivity, lbrow_connectivity, rbrow_connectivity,
        nose_connectivity, leye_connectivity, reye_connectivity,
        mouth_connectivity])

    # Ignore the two inner mouth points
    ind = np.hstack((np.arange(60), np.arange(61, 64), np.arange(65, 68)))
    new_landmark_group = LandmarkGroup(
        PointUndirectedGraph(landmark_group.lms.points[ind],
                             total_conn),
        OrderedDict([('all', np.ones(n_points, dtype=np.bool))]))

    new_landmark_group['jaw'] = jaw_indices
    new_landmark_group['left_eyebrow'] = lbrow_indices
    new_landmark_group['right_eyebrow'] = rbrow_indices
    new_landmark_group['nose'] = np.hstack([upper_nose_indices,
                                            lower_nose_indices])
    new_landmark_group['left_eye'] = leye_indices
    new_landmark_group['right_eye'] = reye_indices
    new_landmark_group['mouth'] = np.hstack([outer_mouth_indices,
                                             inner_mouth_indices])

    del new_landmark_group['all']  # Remove pointless all group

    return group, new_landmark_group


def ibug_face_51(landmark_group):
    """
    Apply the ibug's "standard" 51 point semantic labels (based on the
    original semantic labels of multiPIE but removing the annotations
    corresponding to the jaw region) to the landmark group.

    The group label will be ``ibug_face_51``.

    The semantic labels applied are as follows:

      - left_eyebrow
      - right_eyebrow
      - nose
      - left_eye
      - right_eye
      - mouth

    Parameters
    ----------
    landmark_group : :map:`LandmarkGroup`
        The landmark group to apply semantic labels to.

    Returns
    -------
    group : `str`
        The group label: ``ibug_face_51``
    landmark_group : :map:`LandmarkGroup`
        New landmark group.

    Raises
    ------
    error : :map:`LabellingError`
        If the given landmark group contains less than 68 points

    References
    ----------
    .. [1] http://www.multipie.org/
    """
    from menpo.shape import PointUndirectedGraph

    group = 'ibug_face_51'
    n_points = 51
    _validate_input(landmark_group, 68, group)

    lbrow_indices = np.arange(0, 5)
    rbrow_indices = np.arange(5, 10)
    upper_nose_indices = np.arange(10, 14)
    lower_nose_indices = np.arange(14, 19)
    leye_indices = np.arange(19, 25)
    reye_indices = np.arange(25, 31)
    outer_mouth_indices = np.arange(31, 43)
    inner_mouth_indices = np.arange(43, 51)

    lbrow_connectivity = _connectivity_from_array(lbrow_indices)
    rbrow_connectivity = _connectivity_from_array(rbrow_indices)
    nose_connectivity = np.vstack([
        _connectivity_from_array(upper_nose_indices),
        _connectivity_from_array(lower_nose_indices)])
    leye_connectivity = _connectivity_from_array(leye_indices, close_loop=True)
    reye_connectivity = _connectivity_from_array(reye_indices, close_loop=True)
    mouth_connectivity = np.vstack([
        _connectivity_from_array(outer_mouth_indices, close_loop=True),
        _connectivity_from_array(inner_mouth_indices, close_loop=True)])

    total_conn = np.vstack([
        lbrow_connectivity, rbrow_connectivity, nose_connectivity,
        leye_connectivity, reye_connectivity, mouth_connectivity])

    # Ignore the two inner mouth points
    ind = np.arange(17, 68)
    new_landmark_group = LandmarkGroup(
        PointUndirectedGraph(landmark_group.lms.points[ind],
                             total_conn),
        OrderedDict([('all', np.ones(n_points, dtype=np.bool))]))

    new_landmark_group['left_eyebrow'] = lbrow_indices
    new_landmark_group['right_eyebrow'] = rbrow_indices
    new_landmark_group['nose'] = np.hstack([upper_nose_indices,
                                            lower_nose_indices])
    new_landmark_group['left_eye'] = leye_indices
    new_landmark_group['right_eye'] = reye_indices
    new_landmark_group['mouth'] = np.hstack([outer_mouth_indices,
                                             inner_mouth_indices])

    del new_landmark_group['all']  # Remove pointless all group

    return group, new_landmark_group


def ibug_face_49(landmark_group):
    """
    Apply the ibug's "standard" 49 point semantic labels (based on the
    original semantic labels of multiPIE but removing the annotations
    corresponding to the jaw region and the 2 describing the inner mouth
    corners) to the landmark group.

    The group label will be ``ibug_face_49``.

    The semantic labels applied are as follows:

      - left_eyebrow
      - right_eyebrow
      - nose
      - left_eye
      - right_eye
      - mouth

    Parameters
    ----------
    landmark_group : :map:`LandmarkGroup`
        The landmark group to apply semantic labels to.

    Returns
    -------
    group : `str`
        The group label: ``ibug_face_49``
    landmark_group : :map:`LandmarkGroup`
        New landmark group.

    Raises
    ------
    error : :map:`LabellingError`
        If the given landmark group contains less than 68 points

    References
    ----------
    .. [1] http://www.multipie.org/
    """
    from menpo.shape import PointUndirectedGraph

    group = 'ibug_face_49'
    n_points = 49
    _validate_input(landmark_group, 68, group)

    lbrow_indices = np.arange(0, 5)
    rbrow_indices = np.arange(5, 10)
    upper_nose_indices = np.arange(10, 14)
    lower_nose_indices = np.arange(14, 19)
    leye_indices = np.arange(19, 25)
    reye_indices = np.arange(25, 31)
    outer_mouth_indices = np.arange(31, 43)
    inner_mouth_indices = np.hstack((31, np.arange(43, 46),
                                     37, np.arange(46, 49)))

    lbrow_connectivity = _connectivity_from_array(lbrow_indices)
    rbrow_connectivity = _connectivity_from_array(rbrow_indices)
    nose_connectivity = np.vstack([
        _connectivity_from_array(upper_nose_indices),
        _connectivity_from_array(lower_nose_indices)])
    leye_connectivity = _connectivity_from_array(leye_indices, close_loop=True)
    reye_connectivity = _connectivity_from_array(reye_indices, close_loop=True)
    mouth_connectivity = np.vstack([
        _connectivity_from_array(outer_mouth_indices, close_loop=True),
        _connectivity_from_array(inner_mouth_indices, close_loop=True)])

    total_conn = np.vstack([
        lbrow_connectivity, rbrow_connectivity, nose_connectivity,
        leye_connectivity, reye_connectivity, mouth_connectivity])

    # Ignore the two inner mouth points
    ind = np.hstack((np.arange(17, 60), np.arange(61, 64), np.arange(65, 68)))
    new_landmark_group = LandmarkGroup(
        PointUndirectedGraph(landmark_group.lms.points[ind],
                             total_conn),
        OrderedDict([('all', np.ones(n_points, dtype=np.bool))]))

    new_landmark_group['left_eyebrow'] = lbrow_indices
    new_landmark_group['right_eyebrow'] = rbrow_indices
    new_landmark_group['nose'] = np.hstack([upper_nose_indices,
                                            lower_nose_indices])
    new_landmark_group['left_eye'] = leye_indices
    new_landmark_group['right_eye'] = reye_indices
    new_landmark_group['mouth'] = np.hstack([outer_mouth_indices,
                                             inner_mouth_indices])

    del new_landmark_group['all']  # Remove pointless all group

    return group, new_landmark_group


def ibug_face_68_trimesh(landmark_group):
    """
    Apply the ibug's "standard" 68 point triangulation to the landmarks in
    the given landmark group.

    The group label will be ``ibug_face_68_trimesh``.

    The semantic labels applied are as follows:

      - tri

    Parameters
    ----------
    landmark_group : :map:`LandmarkGroup`
        The landmark group to apply semantic labels to.

    Returns
    -------
    group : `str`
        The group label: ``ibug_face_68_trimesh``
    landmark_group : :map:`LandmarkGroup`
        New landmark group.

    Raises
    ------
    error : :map:`LabellingError`
        If the given landmark group contains less than 68 points

    References
    ----------
    .. [1] http://www.multipie.org/
    """
    from menpo.shape import TriMesh

    group = 'ibug_face_68_trimesh'
    n_expected_points = 68
    n_points = landmark_group.lms.n_points

    _validate_input(landmark_group, n_expected_points, group)

    tri_list = np.array([[47, 29, 28], [44, 43, 23], [38, 20, 21], [47, 28,42],
                        [49, 61, 60], [40, 41, 37], [37, 19, 20], [28, 40, 39],
                        [38, 21, 39], [36,  1, 	0], [48, 59,  4], [49, 60, 48],
                        [67, 59, 60], [13, 53, 14], [61, 51, 62], [57,  8,  7],
                        [52, 51, 33], [61, 67, 60], [52, 63, 51], [66, 56, 57],
                        [35, 30, 29], [53, 52, 35], [37, 36, 17], [18, 37, 17],
                        [37, 38, 40], [38, 37, 20], [19, 37, 18], [38, 39, 40],
                        [28, 29, 40], [41, 36, 37], [27, 39, 21], [41, 31,  1],
                        [30, 32, 31], [33, 51, 50], [33, 30, 34], [31, 40, 29],
                        [36,  0, 17], [31,  2,  1], [31, 41, 40], [ 1, 36, 41],
                        [31, 49,  2], [ 2, 49,  3], [60, 59, 48], [ 3, 49, 48],
                        [31, 32, 50], [48,  4,  3], [59,  5,  4], [58, 67, 66],
                        [ 5, 59, 58], [58, 59, 67], [ 7,  6, 58], [66, 57, 58],
                        [13, 54, 53], [ 7, 58, 57], [ 6,  5, 58], [50, 61, 49],
                        [62, 67, 61], [31, 50, 49], [32, 33, 50], [30, 33, 32],
                        [34, 52, 33], [35, 52, 34], [53, 63, 52], [62, 63, 65],
                        [62, 51, 63], [66, 65, 56], [63, 53, 64], [62, 66, 67],
                        [62, 65, 66], [57, 56,  9], [65, 63, 64], [ 8, 57,  9],
                        [ 9, 56, 10], [10, 56, 11], [11, 56, 55], [11, 55, 12],
                        [56, 65, 55], [55, 64, 54], [55, 65, 64], [55, 54, 12],
                        [64, 53, 54], [12, 54, 13], [45, 46, 44], [35, 34, 30],
                        [14, 53, 35], [15, 46, 45], [27, 28, 39], [27, 42, 28],
                        [35, 29, 47], [30, 31, 29], [15, 35, 46], [15, 14, 35],
                        [43, 22, 23], [27, 21, 22], [24, 44, 23], [44, 47, 43],
                        [43, 47, 42], [46, 35, 47], [26, 45, 44], [46, 47, 44],
                        [25, 44, 24], [25, 26, 44], [16, 15, 45], [16, 45, 26],
                        [22, 42, 43], [50, 51, 61], [27, 22, 42]])
    new_landmark_group = LandmarkGroup(
        TriMesh(landmark_group.lms.points, tri_list, copy=False),
        OrderedDict([('tri', np.ones(n_points, dtype=np.bool))]))

    return group, new_landmark_group


def ibug_face_65_closed_mouth(landmark_group):
    """
    Apply the ibug's "standard" 68 point semantic labels (based on the
    original semantic labels of multiPIE) to the landmarks in
    the given landmark group - but ignore the 3 points that are coincident for
    a closed mouth. Therefore, there only 65 points are returned.

    The group label will be ``ibug_face_65_closed_mouth``.

    The semantic labels applied are as follows:

      - jaw
      - left_eyebrow
      - right_eyebrow
      - nose
      - left_eye
      - right_eye
      - mouth

    Parameters
    ----------
    landmark_group : :map:`LandmarkGroup`
        The landmark group to apply semantic labels to.

    Returns
    -------
    group : `str`
        The group label: ``ibug_face_65_closed_mouth``
    landmark_group : :map:`LandmarkGroup`
        New landmark group.

    Raises
    ------
    error : :map:`LabellingError`
        If the given landmark group contains less than 68 points

    References
    ----------
    .. [1] http://www.multipie.org/
    """
    from menpo.shape import PointUndirectedGraph

    group = 'ibug_face_65_closed_mouth'
    n_points = 65
    _validate_input(landmark_group, 68, group)

    jaw_indices = np.arange(0, 17)
    lbrow_indices = np.arange(17, 22)
    rbrow_indices = np.arange(22, 27)
    upper_nose_indices = np.arange(27, 31)
    lower_nose_indices = np.arange(31, 36)
    leye_indices = np.arange(36, 42)
    reye_indices = np.arange(42, 48)
    outer_mouth_indices = np.arange(48, 60)
    inner_mouth_indices = np.arange(60, 65)

    jaw_connectivity = _connectivity_from_array(jaw_indices)
    lbrow_connectivity = _connectivity_from_array(lbrow_indices)
    rbrow_connectivity = _connectivity_from_array(rbrow_indices)
    nose_connectivity = np.vstack([
        _connectivity_from_array(upper_nose_indices),
        _connectivity_from_array(lower_nose_indices)])
    leye_connectivity = _connectivity_from_array(leye_indices, close_loop=True)
    reye_connectivity = _connectivity_from_array(reye_indices, close_loop=True)
    mouth_connectivity = np.vstack([
        _connectivity_from_array(outer_mouth_indices, close_loop=True),
        _connectivity_from_array(inner_mouth_indices)])

    total_conn = np.vstack([
        jaw_connectivity, lbrow_connectivity, rbrow_connectivity,
        nose_connectivity, leye_connectivity, reye_connectivity,
        mouth_connectivity])

    # Ignore the two inner mouth points
    ind = np.arange(65)
    new_landmark_group = LandmarkGroup(
        PointUndirectedGraph(landmark_group.lms.points[ind],
                             total_conn),
        OrderedDict([('all', np.ones(n_points, dtype=np.bool))]))

    new_landmark_group['jaw'] = jaw_indices
    new_landmark_group['left_eyebrow'] = lbrow_indices
    new_landmark_group['right_eyebrow'] = rbrow_indices
    new_landmark_group['nose'] = np.hstack([upper_nose_indices,
                                            lower_nose_indices])
    new_landmark_group['left_eye'] = leye_indices
    new_landmark_group['right_eye'] = reye_indices
    new_landmark_group['mouth'] = np.hstack([outer_mouth_indices,
                                             inner_mouth_indices])
    del new_landmark_group['all']  # Remove pointless all group

    return group, new_landmark_group


def imm_face(landmark_group):
    """
    Apply the 58 point semantic labels from the IMM dataset to the
    landmarks in the given landmark group.

    The group label will be ``imm_face``.

    The semantic labels applied are as follows:

      - jaw
      - left_eye
      - right_eye
      - left_eyebrow
      - right_eyebrow
      - mouth
      - nose

    Parameters
    ----------
    landmark_group : :map:`LandmarkGroup`
        The landmark group to apply semantic labels to.

    Returns
    -------
    group : `str`
        The group label: ``imm_face``
    landmark_group : :map:`LandmarkGroup`
        New landmark group

    Raises
    ------
    error : :map:`LabellingError`
        If the given landmark group contains less than 58 points

    References
    ----------
    .. [1] http://www2.imm.dtu.dk/~aam/
    """
    group = 'imm_face'
    _validate_input(landmark_group, 58, group)
    labels = OrderedDict([
        ('jaw', (0, 13, False)),
        ('left_eye', (13, 21, True)),
        ('right_eye', (21, 29, True)),
        ('left _eyebrow', (29, 34, False)),
        ('right_eyebrow', (34, 39, False)),
        ('mouth', (39, 47, True)),
        ('nose', (47, 58, False))
    ])
    return group, _relabel_group_from_dict(landmark_group.lms, labels)


def lfpw_face(landmark_group):
    """
    Apply the 29 point semantic labels from the LFPW dataset to the
    landmarks in the given landmark group.

    The group label will be ``lfpw_face``.

    The semantic labels applied are as follows:

      - chin
      - left_eye
      - right_eye
      - left_eyebrow
      - right_eyebrow
      - mouth
      - nose

    Parameters
    ----------
    landmark_group: :class:`menpo.landmark.base.LandmarkGroup`
        The landmark group to apply semantic labels to.

    Returns
    -------
    group : `str`
        The group label: ``lfpw_face``
    landmark_group: :map:`LandmarkGroup`
        New landmark group

    Raises
    ------
    error : :map:`LabellingError`
        If the given landmark group contains less than 29 points

    References
    ----------
    .. [1] http://homes.cs.washington.edu/~neeraj/databases/lfpw/
    """
    from menpo.shape import PointUndirectedGraph

    group = 'lfpw_face'
    n_points = 29
    _validate_input(landmark_group, 29, group)

    chin_indices = np.array([28])
    outer_leye_indices = np.array([8, 12, 10, 13])
    pupil_leye_indices = np.array([16])
    outer_reye_indices = np.array([11, 14, 9, 15])
    pupil_reye_indices = np.array([17])
    lbrow_indices = np.array([0, 4, 2, 5])
    rbrow_indices = np.array([3, 6, 1, 7])
    outer_mouth_indices = np.array([22, 24, 23, 27])
    inner_mouth_indices = np.array([22, 25, 23, 26])
    nose_indices = np.array([18, 20, 19, 21])

    # TODO: Not sure this makes a lot of sense...
    chin_connectivity = _connectivity_from_array(chin_indices, close_loop=True)
    leye_connectivity = _connectivity_from_array(outer_leye_indices,
                                                 close_loop=True)
    reye_connectivity = _connectivity_from_array(outer_reye_indices,
                                                 close_loop=True)
    lbrow_connectivity = _connectivity_from_array(lbrow_indices,
                                                  close_loop=True)
    rbrow_connectivity = _connectivity_from_array(rbrow_indices,
                                                  close_loop=True)
    mouth_connectivity = np.vstack(
        (_connectivity_from_array(outer_mouth_indices, close_loop=True),
         _connectivity_from_array(inner_mouth_indices)))
    nose_connectivity = _connectivity_from_array(nose_indices, close_loop=True)

    total_conn = np.vstack(
        (chin_connectivity, leye_connectivity, reye_connectivity,
         lbrow_connectivity, rbrow_connectivity, mouth_connectivity,
         nose_connectivity))

    # Ignore the two inner mouth points
    ind = np.arange(29)
    new_landmark_group = LandmarkGroup(
        PointUndirectedGraph(landmark_group.lms.points[ind],
                             total_conn),
        OrderedDict([('all', np.ones(n_points, dtype=np.bool))]))

    new_landmark_group['chin'] = chin_indices
    new_landmark_group['left_eye'] = np.hstack((outer_leye_indices,
                                                pupil_leye_indices))
    new_landmark_group['right_eye'] = np.hstack((outer_reye_indices,
                                                 pupil_reye_indices))
    new_landmark_group['left_eyebrow'] = lbrow_indices
    new_landmark_group['right_eyebrow'] = rbrow_indices
    new_landmark_group['mouth'] = np.hstack((outer_mouth_indices,
                                             inner_mouth_indices))
    new_landmark_group['nose'] = nose_indices
    del new_landmark_group['all']  # Remove pointless all group

    return group, new_landmark_group


def _build_upper_eyelid():
    top_indices = np.arange(0, 7)
    middle_indices = np.arange(12, 17)
    upper_eyelid_indices = np.hstack((top_indices, middle_indices))

    upper_eyelid_connectivity = zip(top_indices, top_indices[1:])
    upper_eyelid_connectivity += [(0, 12)]
    upper_eyelid_connectivity += zip(middle_indices, middle_indices[1:])
    upper_eyelid_connectivity += [(16, 6)]

    return upper_eyelid_indices, upper_eyelid_connectivity


def ibug_open_eye(landmark_group):
    """
    Apply the ibug's "standard" open eye semantic labels to the
    landmarks in the given landmark group.

    The group label will be ``ibug_open_eye``.

    The semantic labels applied are as follows:

      - upper_eyelid
      - lower_eyelid
      - iris
      - pupil
      - sclera

    Parameters
    ----------
    landmark_group : :map:`LandmarkGroup`
        The landmark group to apply semantic labels to.

    Returns
    -------
    group : `str`
        The group label: ``ibug_open_eye``
    landmark_group : :map:`LandmarkGroup`
        New landmark group.

    Raises
    ------
    error : :map:`LabellingError`
        If the given landmark group contains less than 38 points
    """
    from menpo.shape import PointUndirectedGraph

    group = 'ibug_open_eye'
    n_expected_points = 38
    n_points = landmark_group.lms.n_points

    _validate_input(landmark_group, n_expected_points, group)

    upper_el_indices, upper_el_connectivity = _build_upper_eyelid()

    iris_range = (22, 30)
    pupil_range = (30, 38)
    sclera_top = np.arange(12, 17)
    sclera_bottom = np.arange(17, 22)
    sclera_indices = np.hstack((0, sclera_top, 6, sclera_bottom))
    lower_el_top = np.arange(17, 22)
    lower_el_bottom = np.arange(7, 12)
    lower_el_indices = np.hstack((6, lower_el_top, 0, lower_el_bottom))

    iris_connectivity = _connectivity_from_range(iris_range, close_loop=True)
    pupil_connectivity = _connectivity_from_range(pupil_range, close_loop=True)

    sclera_connectivity = zip(sclera_top, sclera_top[1:])
    sclera_connectivity += [(0, 21)]
    sclera_connectivity += zip(sclera_bottom, sclera_bottom[1:])
    sclera_connectivity += [(6, 17)]

    lower_el_connectivity = zip(lower_el_top, lower_el_top[1:])
    lower_el_connectivity += [(6, 7)]
    lower_el_connectivity += zip(lower_el_bottom, lower_el_bottom[1:])
    lower_el_connectivity += [(11, 0)]

    total_conn = np.asarray(upper_el_connectivity +
                            lower_el_connectivity +
                            iris_connectivity.tolist() +
                            pupil_connectivity.tolist() +
                            sclera_connectivity)
    new_landmark_group = LandmarkGroup(
        PointUndirectedGraph(landmark_group.lms.points, total_conn),
        OrderedDict([('all', np.ones(n_points, dtype=np.bool))]))

    new_landmark_group['upper_eyelid'] = upper_el_indices
    new_landmark_group['lower_eyelid'] = lower_el_indices
    new_landmark_group['pupil'] = np.arange(*pupil_range)
    new_landmark_group['iris'] = np.arange(*iris_range)
    new_landmark_group['sclera'] = sclera_indices
    del new_landmark_group['all']  # Remove pointless all group

    return group, new_landmark_group


def ibug_close_eye_points(landmark_group):
    """
    Apply the ibug's "standard" close eye semantic labels to the
    landmarks in the given landmark group.

    The group label will be ``ibug_close_eye``.

    The semantic labels applied are as follows:

      - upper_eyelid
      - lower_eyelid

    Parameters
    ----------
    landmark_group : :map:`LandmarkGroup`
        The landmark group to apply semantic labels to.

    Returns
    -------
    group : `str`
        The group label: ``ibug_close_eye``
    landmark_group : :map:`LandmarkGroup`
        New landmark group.

    Raises
    ------
    error : :map:`LabellingError`
        If the given landmark group contains less than 17 points
    """
    from menpo.shape import PointUndirectedGraph

    group = 'ibug_close_eye'
    n_expected_points = 17
    n_points = landmark_group.lms.n_points
    _validate_input(landmark_group, n_expected_points, group)

    upper_indices, upper_connectivity = _build_upper_eyelid()

    middle_indices = np.arange(12, 17)
    bottom_indices = np.arange(6, 12)
    lower_indices = np.hstack((bottom_indices, 0, middle_indices))
    lower_connectivity = zip(bottom_indices, bottom_indices[1:])
    lower_connectivity += [(0, 12)]
    lower_connectivity += zip(middle_indices, middle_indices[1:])
    lower_connectivity += [(11, 0)]

    total_conn = np.asarray(upper_connectivity + lower_connectivity)
    new_landmark_group = LandmarkGroup(
        PointUndirectedGraph(landmark_group.lms.points, total_conn),
        OrderedDict([('all', np.ones(n_points, dtype=np.bool))]))

    new_landmark_group['upper_eyelid'] = upper_indices
    new_landmark_group['lower_eyelid'] = lower_indices
    del new_landmark_group['all']  # Remove pointless all group

    return group, new_landmark_group


def ibug_open_eye_trimesh(landmark_group):
    """
    Apply the ibug's "standard" open eye semantic labels to the
    landmarks in the given landmark group.

    The group label will be ``ibug_open_eye_trimesh``.

    The semantic labels applied are as follows:

      - tri

    Parameters
    ----------
    landmark_group : :map:`LandmarkGroup`
        The landmark group to apply semantic labels to.

    Returns
    -------
    group : `str`
        The group label: ``ibug_open_eye_trimesh``
    landmark_group : :map:`LandmarkGroup`
        New landmark group.

    Raises
    ------
    error : :map:`LabellingError`
        If the given landmark group contains less than 38 points
    """
    from menpo.shape import TriMesh

    group = 'ibug_open_eye_trimesh'
    n_expected_points = 38
    n_points = landmark_group.lms.n_points

    _validate_input(landmark_group, n_expected_points, group)

    tri_list = np.array([[29, 36, 28], [22, 13, 23], [12,  1,  2],
                         [29, 30, 37], [13,  3, 14], [13, 12,  2],
                         [19,  8,  9], [25, 33, 24], [36, 37, 33],
                         [24, 32, 31], [33, 37, 31], [35, 34, 27],
                         [35, 36, 33], [ 3, 13,  2], [14, 24, 23],
                         [33, 32, 24], [15, 25, 14], [25, 26, 34],
                         [22, 30, 29], [31, 37, 30], [24, 31, 23],
                         [32, 33, 31], [22, 12, 13], [ 0,  1, 12],
                         [14, 23, 13], [31, 30, 23], [28, 19, 20],
                         [21, 11,  0], [12, 21,  0], [20, 11, 21],
                         [20, 10, 11], [21, 29, 20], [21, 12, 22],
                         [30, 22, 23], [29, 21, 22], [27, 19, 28],
                         [29, 37, 36], [29, 28, 20], [36, 35, 28],
                         [20, 19, 10], [10, 19,  9], [28, 35, 27],
                         [19, 19,  8], [17, 16,  6], [18,  7,  8],
                         [25, 34, 33], [18, 27, 17], [18, 19, 27],
                         [18, 17,  7], [27, 26, 17], [17,  6,  7],
                         [14, 25, 24], [34, 35, 33], [17, 26, 16],
                         [27, 34, 26], [ 3, 15, 14], [15, 26, 25],
                         [ 4, 15,  3], [16, 26, 15], [16,  4,  5],
                         [16, 15,  4], [16,  5,  6], [8, 18, 19]])

    new_landmark_group = LandmarkGroup(
        TriMesh(landmark_group.lms.points, tri_list, copy=False),
        OrderedDict([('tri', np.ones(n_points, dtype=np.bool))]))

    return group, new_landmark_group


def ibug_close_eye_trimesh(landmark_group):
    """
    Apply the ibug's "standard" close eye semantic labels to the
    landmarks in the given landmark group.

    The group label will be ``ibug_close_eye_trimesh``.

    The semantic labels applied are as follows:

      - tri

    Parameters
    ----------
    landmark_group : :map:`LandmarkGroup`
        The landmark group to apply semantic labels to.

    Returns
    -------
    group : `str`
        The group label: ``ibug_close_eye_trimesh``
    landmark_group : :map:`LandmarkGroup`
        New landmark group.

    Raises
    ------
    error : :map:`LabellingError`
        If the given landmark group contains less than 38 points
    """
    from menpo.shape import TriMesh

    group = 'ibug_close_eye_trimesh'
    n_expected_points = 17
    n_points = landmark_group.lms.n_points

    _validate_input(landmark_group, n_expected_points, group)

    tri_list = np.array([[10, 11, 13], [ 3, 13,  2], [ 4, 14,  3],
                         [15,  5, 16], [12, 11,  0], [13, 14, 10],
                         [13, 12,  2], [14, 13,  3], [ 0,  1, 12],
                         [ 2, 12,  1], [13, 11, 12], [ 9, 10, 14],
                         [15,  9, 14], [ 7,  8, 15], [ 5,  6, 16],
                         [15, 14,  4], [ 7, 15, 16], [ 8,  9, 15],
                         [15,  4,  5], [16,  6,  7]])

    new_landmark_group = LandmarkGroup(
        TriMesh(landmark_group.lms.points, tri_list, copy=False),
        OrderedDict([('tri', np.ones(n_points, dtype=np.bool))]))

    return group, new_landmark_group


def ibug_tongue(landmark_group):
    """
    Apply the ibug's "standard" tongue semantic labels to the landmarks in the
    given landmark group.

    The group label will be ``ibug_tongue``.

    The semantic labels applied are as follows:

      - outline
      - bisector

    Parameters
    ----------
    landmark_group : :map:`LandmarkGroup`
        The landmark group to apply semantic labels to.

    Returns
    -------
    group : `str`
        The group label: ``ibug_tongue``
    landmark_group : :map:`LandmarkGroup`
        New landmark group.

    Raises
    ------
    error : :map:`LabellingError`
        If the given landmark group contains less than 19 points
    """
    group = 'ibug_tongue'
    _validate_input(landmark_group, 19, group)
    labels = OrderedDict([
        ('outline', (0, 13, False)),
        ('bisector', (13, 19, False))
    ])
    return group, _relabel_group_from_dict(landmark_group.lms, labels)


def ibug_hand(landmark_group):
    """
    Apply the ibug's "standard" 39 point semantic labels to the landmark group.

    The group label will be ``ibug_hand``.

    The semantic labels applied are as follows:

      - thumb
      - index
      - middle
      - ring
      - pinky
      - palm

    Parameters
    ----------
    landmark_group : :map:`LandmarkGroup`
        The landmark group to apply semantic labels to.

    Returns
    -------
    group : `str`
        The group label: ``ibug_hand``
    landmark_group : :map:`LandmarkGroup`
        New landmark group.

    Raises
    ------
    error : :map:`LabellingError`
        If the given landmark group contains less than 39 points
    """
    from menpo.shape import PointUndirectedGraph

    group = 'ibug_hand'
    n_points = landmark_group.lms.n_points
    _validate_input(landmark_group, 39, group)

    thumb_indices = np.arange(0, 5)
    index_indices = np.arange(5, 12)
    middle_indices = np.arange(12, 19)
    ring_indices = np.arange(19, 26)
    pinky_indices = np.arange(26, 33)
    palm_indices = np.hstack((np.array([32, 25, 18, 11, 33, 34, 4]),
                              np.arange(35, 39)))

    thumb_connectivity = _connectivity_from_array(thumb_indices,
                                                  close_loop=False)
    index_connectivity = _connectivity_from_array(index_indices,
                                                  close_loop=False)
    middle_connectivity = _connectivity_from_array(middle_indices,
                                                   close_loop=False)
    ring_connectivity = _connectivity_from_array(ring_indices,
                                                 close_loop=False)
    pinky_connectivity = _connectivity_from_array(pinky_indices,
                                                  close_loop=False)
    palm_connectivity = _connectivity_from_array(palm_indices,
                                                 close_loop=True)

    total_conn = np.vstack((thumb_connectivity, index_connectivity,
                            middle_connectivity, ring_connectivity,
                            pinky_connectivity, palm_connectivity))

    new_landmark_group = LandmarkGroup(
        PointUndirectedGraph(landmark_group.lms.points, total_conn),
        OrderedDict([('all', np.ones(n_points, dtype=np.bool))]))

    new_landmark_group['thumb'] = thumb_indices
    new_landmark_group['index'] = index_indices
    new_landmark_group['middle'] = middle_indices
    new_landmark_group['ring'] = ring_indices
    new_landmark_group['pinky'] = pinky_indices
    new_landmark_group['palm'] = palm_indices
    del new_landmark_group['all']  # Remove pointless all group

    return group, new_landmark_group


def stickmen_pose(landmark_group):
    """
    Apply the stickmen "standard" 12 point semantic labels to the landmark
    group.

    The group label will be ``stickmen_pose``.

    The semantic labels applied are as follows:

      - torso
      - right_upper_arm
      - left_upper_arm
      - right_lower_arm
      - left_lower_arm
      - head

    Parameters
    ----------
    landmark_group : :map:`LandmarkGroup`
        The landmark group to apply semantic labels to.

    Returns
    -------
    group : `str`
        The group label: ``stickmen_pose``
    landmark_group : :map:`LandmarkGroup`
        New landmark group.

    Raises
    ------
    error : :map:`LabellingError`
        If the given landmark group contains less than 12 points

    References
    ----------
    .. [1] http://www.robots.ox.ac.uk/~vgg/data/stickmen/
    """
    group = 'stickmen_pose'
    _validate_input(landmark_group, 12, group)
    labels = OrderedDict([
        ('torso', (0, 2, False)),
        ('right_upper arm', (2, 4, False)),
        ('left_upper arm', (4, 6, False)),
        ('right_lower_arm', (6, 8, False)),
        ('left_lower_arm', (8, 10, False)),
        ('head', (10, 12, False))
    ])
    return group, _relabel_group_from_dict(landmark_group.lms, labels)


def lsp_pose(landmark_group):
    """
    Apply the lsp "standard" 14 point semantic labels to the landmark
    group.

    The group label will be ``lsp_pose``.

    The semantic labels applied are as follows:

      - left_leg
      - right_leg
      - left_arm
      - right_arm
      - head

    Parameters
    ----------
    landmark_group : :map:`LandmarkGroup`
        The landmark group to apply semantic labels to.

    Returns
    -------
    group : `str`
        The group label: ``lsp_pose``
    landmark_group : :map:`LandmarkGroup`
        New landmark group.

    Raises
    ------
    error : :map:`LabellingError`
        If the given landmark group contains less than 14 points

    References
    ----------
    .. [1] http://www.comp.leeds.ac.uk/mat4saj/lsp.html
    """
    from menpo.shape import PointUndirectedGraph

    group = 'lsp_pose'
    n_points = landmark_group.lms.n_points
    _validate_input(landmark_group, 14, group)

    left_leg_indices = np.arange(0, 3)
    right_leg_indices = np.arange(3, 6)
    left_arm_indices = np.arange(6, 9)
    right_arm_indices = np.arange(9, 12)
    head_indices = np.arange(12, 14)

    left_leg_connectivity = _connectivity_from_array(left_leg_indices)
    right_leg_connectivity = _connectivity_from_array(right_leg_indices)
    left_arm_connectivity = _connectivity_from_array(left_arm_indices)
    right_arm_connectivity = _connectivity_from_array(right_arm_indices)
    head_connectivity = _connectivity_from_array(head_indices)

    total_conn = np.vstack([left_leg_connectivity,
                            right_leg_connectivity,
                            left_arm_connectivity,
                            right_arm_connectivity,
                            head_connectivity])

    new_landmark_group = LandmarkGroup(
        PointUndirectedGraph(landmark_group.lms.points, total_conn),
        OrderedDict([('all', np.ones(n_points, dtype=np.bool))]))

    new_landmark_group['left_leg'] = left_leg_indices
    new_landmark_group['right_leg'] = right_leg_indices
    new_landmark_group['left_arm'] = left_arm_indices
    new_landmark_group['right_arm'] = right_arm_indices
    new_landmark_group['head'] = head_indices

    del new_landmark_group['all']  # Remove pointless all group

    return group, new_landmark_group


def flic_pose(landmark_group):
    """
    Apply the flic "standard" 11 point semantic labels to the landmark
    group.

    The group label will be ``flic_pose``.

    The semantic labels applied are as follows:

      - left_arm
      - right_arm
      - hips
      - face

    Parameters
    ----------
    landmark_group : :map:`LandmarkGroup`
        The landmark group to apply semantic labels to.

    Returns
    -------
    group : `str`
        The group label: ``flic_pose``
    landmark_group : :map:`LandmarkGroup`
        New landmark group.

    Raises
    ------
    error : :map:`LabellingError`
        If the given landmark group contains less than 11 points

    References
    ----------
    .. [1] http://vision.grasp.upenn.edu/cgi-bin/index.php?n=VideoLearning.FLIC
    """
    group = 'flic_pose'
    _validate_input(landmark_group, 11, group)
    labels = OrderedDict([
        ('left_arm', (0, 3, False)),
        ('right_arm', (3, 6, False)),
        ('hips', (6, 8, False)),
        ('face', (8, 11, True))])

    return group, _relabel_group_from_dict(landmark_group.lms, labels)


def streetscene_car_view_0(landmark_group):
    """
    Apply the 8 point semantic labels of the view 0  of the MIT Street Scene
    Car dataset to the landmark group.

    The group label will be ``streetscene_car_view_0``.

    The semantic labels applied are as follows:

      - front
      - bonnet
      - windshield

    Parameters
    ----------
    landmark_group : :map:`LandmarkGroup`
        The landmark group to apply semantic labels to.

    Returns
    -------
    group : `str`
        The group label: ``streetscene_car_view_0``
    landmark_group : :map:`LandmarkGroup`
        New landmark group.

    Raises
    ------
    error : :map:`LabellingError`
        If the given landmark group contains less than 20 points

    References
    ----------
    .. [1] http://www.cs.cmu.edu/~vboddeti/alignment.html
    """
    from menpo.shape import PointUndirectedGraph

    group = 'streetscene_car_view_0'
    n_points = 8
    _validate_input(landmark_group, 20, group)

    front_indices = np.array([0, 1, 3, 2])
    bonnet_indices = np.array([2, 3, 5, 4])
    windshield_indices = np.array([4, 5, 7, 6])

    front_connectivity = _connectivity_from_array(front_indices,
                                                  close_loop=True)
    bonnet_connectivity = _connectivity_from_array(bonnet_indices,
                                                   close_loop=True)
    windshield_connectivity = _connectivity_from_array(windshield_indices,
                                                       close_loop=True)

    total_conn = np.vstack((front_connectivity, bonnet_connectivity,
                            windshield_connectivity))

    ind = np.arange(8)
    new_landmark_group = LandmarkGroup(
        PointUndirectedGraph(landmark_group.lms.points[ind], total_conn),
        OrderedDict([('all', np.ones(n_points, dtype=np.bool))]))

    new_landmark_group['front'] = front_indices
    new_landmark_group['bonnet'] = bonnet_indices
    new_landmark_group['windshield'] = windshield_indices
    del new_landmark_group['all']  # Remove pointless all group

    return group, new_landmark_group


def streetscene_car_view_1(landmark_group):
    """
    Apply the 14 point semantic labels of the view 1  of the MIT Street Scene
    Car dataset to the landmark group.

    The group label will be ``streetscene_car_view_1``.

    The semantic labels applied are as follows:

      - front
      - bonnet
      - windshield
      - left_side

    Parameters
    ----------
    landmark_group : :map:`LandmarkGroup`
        The landmark group to apply semantic labels to.

    Returns
    -------
    group : `str`
        The group label: ``streetscene_car_view_1``
    landmark_group : :map:`LandmarkGroup`
        New landmark group.

    Raises
    ------
    error : :map:`LabellingError`
        If the given landmark group contains less than 20 points

    References
    ----------
    .. [1] http://www.cs.cmu.edu/~vboddeti/alignment.html
    """
    from menpo.shape import PointUndirectedGraph

    group = 'streetscene_car_view_1'
    n_points = 14
    _validate_input(landmark_group, 20, group)

    front_indices = np.array([0, 1, 3, 2])
    bonnet_indices = np.array([2, 3, 5, 4])
    windshield_indices = np.array([4, 5, 7, 6])
    left_side_indices = np.array([0, 2, 4, 6, 8, 9, 10, 11, 13, 12])

    front_connectivity = _connectivity_from_array(front_indices,
                                                  close_loop=True)
    bonnet_connectivity = _connectivity_from_array(bonnet_indices,
                                                   close_loop=True)
    windshield_connectivity = _connectivity_from_array(windshield_indices,
                                                       close_loop=True)
    left_side_connectivity = _connectivity_from_array(left_side_indices,
                                                      close_loop=True)

    total_conn = np.vstack((front_connectivity, bonnet_connectivity,
                            windshield_connectivity, left_side_connectivity))

    ind = np.hstack((np.arange(9), np.array([10, 12, 14, 16, 18])))
    new_landmark_group = LandmarkGroup(
        PointUndirectedGraph(landmark_group.lms.points[ind], total_conn),
        OrderedDict([('all', np.ones(n_points, dtype=np.bool))]))

    new_landmark_group['front'] = front_indices
    new_landmark_group['bonnet'] = bonnet_indices
    new_landmark_group['windshield'] = windshield_indices
    new_landmark_group['left_side'] = left_side_indices
    del new_landmark_group['all']  # Remove pointless all group

    return group, new_landmark_group


def streetscene_car_view_2(landmark_group):
    """
    Apply the 10 point semantic labels of the view 2  of the MIT Street Scene
    Car dataset to the landmark group.

    The group label will be ``streetscene_car_view_2``.

    The semantic labels applied are as follows:

      - left_side

    Parameters
    ----------
    landmark_group : :map:`LandmarkGroup`
        The landmark group to apply semantic labels to.

    Returns
    -------
    group : `str`
        The group label: 'streetscene_car_view_2'
    landmark_group : :map:`LandmarkGroup`
        New landmark group.

    Raises
    ------
    error : :map:`LabellingError`
        If the given landmark group contains less than 20 points

    References
    ----------
    .. [1] http://www.cs.cmu.edu/~vboddeti/alignment.html
    """
    from menpo.shape import PointUndirectedGraph

    group = 'streetscene_car_view_2'
    n_points = 10
    _validate_input(landmark_group, 20, group)

    left_side_indices = np.array([0, 1, 2, 3, 4, 5, 6, 7, 9, 8])

    left_side_connectivity = _connectivity_from_array(left_side_indices,
                                                      close_loop=True)

    total_conn = left_side_connectivity

    ind = np.array([0, 2, 4, 6, 8, 10, 12, 14, 16, 18])
    new_landmark_group = LandmarkGroup(
        PointUndirectedGraph(landmark_group.lms.points[ind], total_conn),
        OrderedDict([('all', np.ones(n_points, dtype=np.bool))]))

    new_landmark_group['left_side'] = left_side_indices
    del new_landmark_group['all']  # Remove pointless all group

    return group, new_landmark_group


def streetscene_car_view_3(landmark_group):
    """
    Apply the 14 point semantic labels of the view 3  of the MIT Street Scene
    Car dataset to the landmark group.

    The group label will be ``streetscene_car_view_2``.

    The semantic labels applied are as follows:

      - left_side
      - rear windshield
      - trunk
      - rear

    Parameters
    ----------
    landmark_group : :map:`LandmarkGroup`
        The landmark group to apply semantic labels to.

    Returns
    -------
    group : `str`
        The group label: ``streetscene_car_view_3``
    landmark_group : :map:`LandmarkGroup`
        New landmark group.

    Raises
    ------
    error : :map:`LabellingError`
        If the given landmark group contains less than 20 points

    References
    ----------
    .. [1] http://www.cs.cmu.edu/~vboddeti/alignment.html
    """
    from menpo.shape import PointUndirectedGraph

    group = 'streetscene_car_view_3'
    n_points = 14
    _validate_input(landmark_group, 20, group)

    left_side_indices = np.array([0, 1, 2, 3, 4, 6, 8, 10, 13, 12])
    rear_windshield_indices = np.array([4, 5, 7, 6])
    trunk_indices = np.array([6, 7, 9, 8])
    rear_indices = np.array([8, 9, 11, 10])

    left_side_connectivity = _connectivity_from_array(left_side_indices,
                                                      close_loop=True)
    rear_windshield_connectivity = _connectivity_from_array(
        rear_windshield_indices, close_loop=True)
    trunk_connectivity = _connectivity_from_array(trunk_indices,
                                                  close_loop=True)
    rear_connectivity = _connectivity_from_array(rear_indices, close_loop=True)

    total_conn = np.vstack((left_side_connectivity,
                            rear_windshield_connectivity,
                            trunk_connectivity, rear_connectivity))

    ind = np.array([0, 2, 4, 6, 8, 9, 10, 11, 12, 13, 14, 15, 16, 18])
    new_landmark_group = LandmarkGroup(
        PointUndirectedGraph(landmark_group.lms.points[ind], total_conn),
        OrderedDict([('all', np.ones(n_points, dtype=np.bool))]))

    new_landmark_group['left_side'] = left_side_indices
    new_landmark_group['rear_windshield'] = rear_windshield_indices
    new_landmark_group['trunk'] = trunk_indices
    new_landmark_group['rear'] = rear_indices
    del new_landmark_group['all']  # Remove pointless all group

    return group, new_landmark_group


def streetscene_car_view_4(landmark_group):
    """
    Apply the 14 point semantic labels of the view 4  of the MIT Street Scene
    Car dataset to the landmark group.

    The group label will be ``streetscene_car_view_4``.

    The semantic labels applied are as follows:

      - front
      - bonnet
      - windshield
      - right_side

    Parameters
    ----------
    landmark_group : :map:`LandmarkGroup`
        The landmark group to apply semantic labels to.

    Returns
    -------
    group : `str`
        The group label: 'streetscene_car_view_4'
    landmark_group : :map:`LandmarkGroup`
        New landmark group.

    Raises
    ------
    error : :map:`LabellingError`
        If the given landmark group contains less than 20 points

    References
    ----------
    .. [1] http://www.cs.cmu.edu/~vboddeti/alignment.html
    """
    from menpo.shape import PointUndirectedGraph

    group = 'streetscene_car_view_4'
    n_points = 14
    _validate_input(landmark_group, 20, group)

    front_indices = np.array([0, 1, 3, 2])
    bonnet_indices = np.array([2, 3, 5, 4])
    windshield_indices = np.array([4, 5, 7, 6])
    right_side_indices = np.array([8, 9, 10, 11, 13, 12, 1, 3, 5, 7])

    front_connectivity = _connectivity_from_array(front_indices,
                                                  close_loop=True)
    bonnet_connectivity = _connectivity_from_array(bonnet_indices,
                                                   close_loop=True)
    windshield_connectivity = _connectivity_from_array(windshield_indices,
                                                       close_loop=True)
    right_side_connectivity = _connectivity_from_array(right_side_indices,
                                                       close_loop=True)

    total_conn = np.vstack((front_connectivity, bonnet_connectivity,
                            windshield_connectivity,
                            right_side_connectivity))

    ind = np.hstack((np.arange(8), np.array([9, 11, 13, 15, 17, 19])))
    new_landmark_group = LandmarkGroup(
        PointUndirectedGraph(landmark_group.lms.points[ind], total_conn),
        OrderedDict([('all', np.ones(n_points, dtype=np.bool))]))

    new_landmark_group['front'] = front_indices
    new_landmark_group['bonnet'] = bonnet_indices
    new_landmark_group['windshield'] = windshield_indices
    new_landmark_group['right_side'] = right_side_indices
    del new_landmark_group['all']  # Remove pointless all group

    return group, new_landmark_group


def streetscene_car_view_5(landmark_group):
    """
    Apply the 10 point semantic labels of the view 5 of the MIT Street Scene
    Car dataset to the landmark group.

    The group label will be ``streetscene_car_view_5``.

    The semantic labels applied are as follows:

      - right_side

    Parameters
    ----------
    landmark_group : :map:`LandmarkGroup`
        The landmark group to apply semantic labels to.

    Returns
    -------
    group : `str`
        The group label: ``streetscene_car_view_5``
    landmark_group : :map:`LandmarkGroup`
        New landmark group.

    Raises
    ------
    error : :map:`LabellingError`
        If the given landmark group contains less than 20 points

    References
    ----------
    .. [1] http://www.cs.cmu.edu/~vboddeti/alignment.html
    """
    from menpo.shape import PointUndirectedGraph

    group = 'streetscene_car_view_5'
    n_points = 10
    _validate_input(landmark_group, 20, group)

    right_side_indices = np.array([0, 1, 2, 3, 4, 5, 6, 7, 9, 8])

    right_side_connectivity = _connectivity_from_array(right_side_indices,
                                                       close_loop=True)

    total_conn = right_side_connectivity

    ind = np.array([1, 3, 5, 7, 9, 11, 13, 15, 17, 19])
    new_landmark_group = LandmarkGroup(
        PointUndirectedGraph(landmark_group.lms.points[ind], total_conn),
        OrderedDict([('all', np.ones(n_points, dtype=np.bool))]))

    new_landmark_group['right_side'] = right_side_indices
    del new_landmark_group['all']  # Remove pointless all group

    return group, new_landmark_group


def streetscene_car_view_6(landmark_group):
    """
    Apply the 14 point semantic labels of the view 6  of the MIT Street Scene
    Car dataset to the landmark group.

    The group label will be ``streetscene_car_view_6``.

    The semantic labels applied are as follows:

      - right_side
      - rear_windshield
      - trunk
      - rear

    Parameters
    ----------
    landmark_group : :map:`LandmarkGroup`
        The landmark group to apply semantic labels to.

    Returns
    -------
    group : `str`
        The group label: ``streetscene_car_view_3``
    landmark_group : :map:`LandmarkGroup`
        New landmark group.

    Raises
    ------
    error : :map:`LabellingError`
        If the given landmark group contains less than 20 points

    References
    ----------
    .. [1] http://www.cs.cmu.edu/~vboddeti/alignment.html
    """
    from menpo.shape import PointUndirectedGraph

    group = 'streetscene_car_view_6'
    n_points = 14
    _validate_input(landmark_group, 20, group)

    right_side_indices = np.array([0, 1, 2, 3, 5, 7, 9, 11, 13, 12])
    rear_windshield_indices = np.array([4, 5, 7, 6])
    trunk_indices = np.array([6, 7, 9, 8])
    rear_indices = np.array([8, 9, 11, 10])

    right_side_connectivity = _connectivity_from_array(right_side_indices,
                                                       close_loop=True)
    rear_windshield_connectivity = _connectivity_from_array(
        rear_windshield_indices, close_loop=True)
    trunk_connectivity = _connectivity_from_array(trunk_indices,
                                                  close_loop=True)
    rear_connectivity = _connectivity_from_array(rear_indices, close_loop=True)

    total_conn = np.vstack((right_side_connectivity,
                            rear_windshield_connectivity,
                            trunk_connectivity, rear_connectivity))

    ind = np.array([1, 3, 5, 7, 8, 9, 10, 11, 12, 13, 14, 15, 17, 19])
    new_landmark_group = LandmarkGroup(
        PointUndirectedGraph(landmark_group.lms.points[ind], total_conn),
        OrderedDict([('all', np.ones(n_points, dtype=np.bool))]))

    new_landmark_group['right_side'] = right_side_indices
    new_landmark_group['rear_windshield'] = rear_windshield_indices
    new_landmark_group['trunk'] = trunk_indices
    new_landmark_group['rear'] = rear_indices
    del new_landmark_group['all']  # Remove pointless all group

    return group, new_landmark_group


def streetscene_car_view_7(landmark_group):
    """
    Apply the 8 point semantic labels of the view 0  of the MIT Street Scene
    Car dataset to the landmark group.

    The group label will be ``streetscene_car_view_7``.

    The semantic labels applied are as follows:

      - rear_windshield
      - trunk
      - rear

    Parameters
    ----------
    landmark_group : :map:`LandmarkGroup`
        The landmark group to apply semantic labels to.

    Returns
    -------
    group : `str`
        The group label: ``streetscene_car_view_7``
    landmark_group : :map:`LandmarkGroup`
        New landmark group.

    Raises
    ------
    error : :map:`LabellingError`
        If the given landmark group contains less than 20 points

    References
    ----------
    .. [1] http://www.cs.cmu.edu/~vboddeti/alignment.html
    """
    from menpo.shape import PointUndirectedGraph

    group = 'streetscene_car_view_7'
    n_points = 8
    _validate_input(landmark_group, 20, group)

    rear_windshield_indices = np.array([0, 1, 3, 2])
    trunk_indices = np.array([2, 3, 5, 4])
    rear_indices = np.array([4, 5, 7, 6])

    rear_windshield_connectivity = _connectivity_from_array(
        rear_windshield_indices, close_loop=True)
    trunk_connectivity = _connectivity_from_array(trunk_indices,
                                                  close_loop=True)
    rear_connectivity = _connectivity_from_array(rear_indices, close_loop=True)

    total_conn = np.vstack((rear_windshield_connectivity,
                            trunk_connectivity, rear_connectivity))

    ind = np.arange(8, 16)
    new_landmark_group = LandmarkGroup(
        PointUndirectedGraph(landmark_group.lms.points[ind], total_conn),
        OrderedDict([('all', np.ones(n_points, dtype=np.bool))]))

    new_landmark_group['rear_windshield'] = rear_windshield_indices
    new_landmark_group['trunk'] = trunk_indices
    new_landmark_group['rear'] = rear_indices
    del new_landmark_group['all']  # Remove pointless all group

    return group, new_landmark_group


def bu3dfe_83(landmark_group):
    """
    Apply the BU-3DFE (Binghamton University 3D Facial Expression)
    Database 83 point facial annotation markup to this landmark group.


    The group label will be ``bu3dfe_83``.

    The semantic labels applied are as follows:

      - right_eye
      - left_eye
      - right_eyebrow
      - left_eyebrow
      - right_nose
      - left_nose
      - nostrils
      - outer_mouth
      - inner_mouth
      - jaw

    Parameters
    ----------
    landmark_group : :map:`LandmarkGroup`
        The landmark group to apply semantic labels to.

    Returns
    -------
    group : `str`
        The group label: ``bu3dfe_83``
    landmark_group : :map:`LandmarkGroup`
        New landmark group.

    Raises
    ------
    :class:`menpo.landmark.exceptions.LabellingError`
        If the given landmark group contains less than 83 points

    References
    ----------
    .. [1] http://www.cs.binghamton.edu/~lijun/Research/3DFE/3DFE_Analysis.html
    """
    from menpo.shape import PointUndirectedGraph

    group = 'bu3dfe_83'
    n_points = 83
    _validate_input(landmark_group, n_points, group)

    reye_indices = np.arange(0, 8)
    leye_indices = np.arange(8, 16)
    rbrow_indices = np.arange(16, 26)
    lbrow_indices = np.arange(26, 36)
    rnose_indicies = np.arange(36, 39)
    lnose_indicies = np.arange(39, 42)
    nostril_indices = np.arange(42, 48)
    outermouth_indices = np.arange(48, 60)
    innermouth_indices = np.arange(60, 68)
    jaw_indices = np.arange(68, 83)

    reye_connectivity = _connectivity_from_array(reye_indices, close_loop=True)
    leye_connectivity = _connectivity_from_array(leye_indices, close_loop=True)
    rbrow_connectivity = _connectivity_from_array(rbrow_indices,
                                                  close_loop=True)
    lbrow_connectivity = _connectivity_from_array(lbrow_indices,
                                                  close_loop=True)
    rnose_connectivity = _connectivity_from_array(rnose_indicies)
    nostril_connectivity = _connectivity_from_array(nostril_indices)
    lnose_connectivity = _connectivity_from_array(lnose_indicies)
    outermouth_connectivity = _connectivity_from_array(outermouth_indices,
                                                       close_loop=True)
    innermouth_connectivity = _connectivity_from_array(innermouth_indices,
                                                       close_loop=True)
    jaw_connectivity = _connectivity_from_array(jaw_indices)

    total_conn = np.vstack([
        reye_connectivity, leye_connectivity,
        rbrow_connectivity, lbrow_connectivity,
        rnose_connectivity, nostril_connectivity, lnose_connectivity,
        outermouth_connectivity, innermouth_connectivity,
        jaw_connectivity
    ])

    new_landmark_group = LandmarkGroup(
        PointUndirectedGraph(landmark_group.lms.points, total_conn),
        OrderedDict([('all', np.ones(n_points, dtype=np.bool))]))

    new_landmark_group['right_eye'] = reye_indices
    new_landmark_group['left_eye'] = leye_indices
    new_landmark_group['right_eyebrow'] = rbrow_indices
    new_landmark_group['left_eyebrow'] = lbrow_indices
    new_landmark_group['right_nose'] = rnose_indicies
    new_landmark_group['left_nose'] = lnose_indicies
    new_landmark_group['nostrils'] = nostril_indices
    new_landmark_group['outer_mouth'] = outermouth_indices
    new_landmark_group['inner_mouth'] = innermouth_indices
    new_landmark_group['jaw'] = jaw_indices

    del new_landmark_group['all']  # Remove pointless all group

    return group, new_landmark_group


def labeller(landmarkable, group, label_func):
    """
    Re-label an existing landmark group on a :map:`Landmarkable` object with a
    new label set.

    Parameters
    ----------
    landmarkable: :map:`Landmarkable`
        :map:`Landmarkable` that will have it's :map:`LandmarkManager`
        augmented with a new :map:`LandmarkGroup`
    group: `str`
        The group label of the existing landmark group that should be
        re-labelled. A copy of this group will be attached to it's landmark
        manager with new labels. The group label of this new group and the
        labels it will have is determined by ``label_func``
    label_func: `func`  -> `(str, LandmarkGroup)`
        A labelling function taken from this module, Takes as input a
        :map:`LandmarkGroup` and returns a tuple of
        (new group label, new LandmarkGroup with semantic labels applied).

    Returns
    -------
    landmarkable : :map:`Landmarkable`
        Augmented ``landmarkable`` (this is just for convenience,
        the object will actually be modified in place)
    """
    new_group, lmark_group = label_func(landmarkable.landmarks[group])
    landmarkable.landmarks[new_group] = lmark_group
    return landmarkable



import warnings
import numpy as np
from menpo.shape import PointCloud
from menpo.testing import is_same_array


def test_pointcloud_creation():
    points = np.array([[1, 2, 3],
                       [1, 1, 1]])
    PointCloud(points)


def test_pointcloud_copy_method():
    points = np.array([[1, 2, 3],
                       [1, 1, 1]])
    landmarks = PointCloud(np.ones([3, 3]), copy=False)

    p = PointCloud(points, copy=False)
    p.landmarks['test'] = landmarks
    p_copy = p.copy()

    assert (not is_same_array(p_copy.points, p.points))
    assert (not is_same_array(p_copy.landmarks['test'].lms.points,
                              p.landmarks['test'].lms.points))


def test_pointcloud_copy_false():
    points = np.array([[1, 2, 3],
                       [1, 1, 1]])
    p = PointCloud(points, copy=False)
    assert (is_same_array(p.points, points))


def test_pointcloud_copy_true():
    points = np.array([[1, 2, 3],
                       [1, 1, 1]])
    p = PointCloud(points)
    assert (not is_same_array(p.points, points))


def test_pointcloud_copy_warning():
    points = np.array([[1, 2, 3],
                       [1, 1, 1]], order='F')
    with warnings.catch_warnings(record=True) as w:
        warnings.simplefilter("always")
        PointCloud(points, copy=False)
        assert len(w) == 1


def test_pointcloud_n_dims():
    points = np.array([[1, 2, 3],
                       [1, 1, 1]])
    pc = PointCloud(points)
    assert (pc.n_dims == 3)


def test_pointcloud_n_points():
    points = np.array([[1, 2, 3],
                       [1, 1, 1]])
    pc = PointCloud(points)
    assert (pc.n_points == 2)


def test_pointcloud_flatten_rebuild():
    points = np.array([[1, 2, 3],
                       [1, 1, 1]])
    pc = PointCloud(points)
    flattened = pc.as_vector()
    new_pc = pc.from_vector(flattened)
    assert (np.all(new_pc.n_dims == pc.n_dims))
    assert (np.all(new_pc.n_points == pc.n_points))
    assert (np.all(pc.points == new_pc.points))



import json
import numpy as np


def LJSONExporter(landmark_group, file_handle):
    r"""
    Given a file handle to write in to (which should act like a Python `file`
    object), write out the landmark data. No value is returned.

    Writes out the LJSON format which is a verbose format that closely
    resembles the landmark group format. It describes semantic labels and
    connectivity between labels. The first axis of the format represents
    the image y-axis and is consistent with ordering within Menpo.

    Parameters
    ----------
    landmark_group : map:`LandmarkGroup`
        The landmark group to write out.
    file_handle : `file`-like object
        The file to write in to
    """
    lg_json = landmark_group.tojson()
    # Add version string
    lg_json['version'] = 2
    return json.dump(lg_json, file_handle, indent=4, separators=(',', ': '),
                     sort_keys=True)


def PTSExporter(landmark_group, file_handle):
    r"""
    Given a file handle to write in to (which should act like a Python `file`
    object), write out the landmark data. No value is returned.

    Writes out the PTS format which is a very simple format that does not
    contain any semantic labels. We assume that the PTS format has been created
    using Matlab and so use 1-based indexing and put the image x-axis as the
    first coordinate (which is the second axis within Menpo).

    Parameters
    ----------
    landmark_group : map:`LandmarkGroup`
        The landmark group to write out.
    file_handle : `file`-like object
        The file to write in to
    """
    pts = landmark_group.lms.points
    # Swap the x and y axis and add 1 to undo our processing
    # We are assuming (as on import) that the landmark file was created using
    # Matlab which is 1 based
    pts = pts[:, [1, 0]] + 1

    header = 'version: 1\nn_points: {}\n{{'.format(pts.shape[0])
    np.savetxt(file_handle, pts, delimiter=' ', header=header, footer='}',
               fmt='%.3f', comments='')



import abc
import os.path

# To debug the Copyable interface, simply uncomment lines 11-23 below and the
# four lines in the copy() method.
# Then you can call print_copyable_log() to see exactly what types have been
# skipped in copying and why.

# from collections import defaultdict
# alien_copies = defaultdict(set)
# non_copies = defaultdict(set)
#
#
# def print_copyable_log():
#     print('Has .copy() but not Copyable:')
#     for k, v in alien_copies.iteritems():
#         print('  {:15}|  {}'.format(k, ', '.join(v)))
#
#     print('\nNo .copy() (shallow copied):')
#     for k, v in non_copies.iteritems():
#         print('  {:15}|  {}'.format(k, ', '.join(v)))


class Copyable(object):
    """
    Efficient copying of classes containing numpy arrays.

    Interface that provides a single method for copying classes very
    efficiently.
    """

    def copy(self):
        r"""
        Generate an efficient copy of this object.

        Note that Numpy arrays and other :map:`Copyable` objects on ``self``
        will be deeply copied. Dictionaries and sets will be shallow copied,
        and everything else will be assigned (no copy will be made).

        Classes that store state other than numpy arrays and immutable types
        should overwrite this method to ensure all state is copied.

        Returns
        -------

        ``type(self)``
            A copy of this object

        """
        # print('copy called on {}'.format(type(self).__name__))
        new = self.__class__.__new__(self.__class__)
        for k, v in self.__dict__.iteritems():
            try:
                new.__dict__[k] = v.copy()
                # if not isinstance(v, Copyable):
                #     alien_copies[type(v).__name__].add(type(self).__name__)
            except AttributeError:
                new.__dict__[k] = v
                # non_copies[type(v).__name__].add(type(self).__name__)
        return new


class Vectorizable(Copyable):
    """
    Flattening of rich objects to vectors and rebuilding them back.

    Interface that provides methods for 'flattening' an object into a
    vector, and restoring from the same vectorized form. Useful for
    statistical analysis of objects, which commonly requires the data
    to be provided as a single vector.
    """

    __metaclass__ = abc.ABCMeta

    @property
    def n_parameters(self):
        r"""The length of the vector that this object produces.

        :type: `int`
        """
        return (self.as_vector()).shape[0]

    def as_vector(self, **kwargs):
        """
        Returns a flattened representation of the object as a single
        vector.

        Returns
        -------
        vector : (N,) ndarray
            The core representation of the object, flattened into a
            single vector. Note that this is always a view back on to the
            original object, but is not writable.
        """
        v = self._as_vector(**kwargs)
        v.flags.writeable = False
        return v

    @abc.abstractmethod
    def _as_vector(self, **kwargs):
        """
        Returns a flattened representation of the object as a single
        vector.

        Returns
        -------
        vector : ``(n_parameters,)`` `ndarray`
            The core representation of the object, flattened into a
            single vector.
        """

    @abc.abstractmethod
    def from_vector_inplace(self, vector):
        """Update the state of this object from a vector form.

        Parameters
        ----------
        vector : ``(n_parameters,)`` `ndarray`
            Flattened representation of this object
        """

    def from_vector(self, vector):
        """Build a new instance of the object from it's vectorized state.

        ``self`` is used to fill out the missing state required to
        rebuild a full object from it's standardized flattened state. This
        is the default implementation, which is which is a ``deepcopy`` of the
        object followed by a call to :meth:`from_vector_inplace()`. This method
        can be overridden for a performance benefit if desired.

        Parameters
        ----------
        vector : ``(n_parameters,)`` `ndarray`
            Flattened representation of the object.

        Returns
        -------
        object : ``type(self)``
            An new instance of this class.
        """
        new = self.copy()
        new.from_vector_inplace(vector)
        return new


class Targetable(Copyable):
    """Interface for objects that can produce a target :map:`PointCloud`.

    This could for instance be the result of an alignment or a generation of a
    :map:`PointCloud` instance from a shape model.

    Implementations must define sensible behavior for:

     - what a target is: see :attr:`target`
     - how to set a target: see :meth:`set_target`
     - how to update the object after a target is set:
       see :meth:`_sync_state_from_target`
     - how to produce a new target after the changes:
       see :meth:`_new_target_from_state`

    Note that :meth:`_sync_target_from_state` needs to be triggered as
    appropriate by subclasses e.g. when :map:`from_vector_inplace` is
    called. This will in turn trigger :meth:`_new_target_from_state`, which each
    subclass must implement.
    """
    __metaclass__ = abc.ABCMeta

    @property
    def n_dims(self):
        r"""The number of dimensions of the :attr:`target`.

        :type: `int`
        """
        return self.target.n_dims

    @property
    def n_points(self):
        r"""The number of points on the :attr:`target`.

        :type: `int`
        """
        return self.target.n_points

    @abc.abstractproperty
    def target(self):
        r"""The current :map:`PointCloud` that this object produces.

        :type: :map:`PointCloud`
        """

    def set_target(self, new_target):
        r"""
        Update this object so that it attempts to recreate the ``new_target``.

        Parameters
        ----------
        new_target : :map:`PointCloud`
            The new target that this object should try and regenerate.
        """
        self._target_setter_with_verification(new_target)  # trigger the update
        self._sync_state_from_target()  # and a sync

    def _target_setter_with_verification(self, new_target):
        r"""Updates the target, checking it is sensible, without triggering a
        sync.

        Should be called by :meth:`_sync_target_from_state` once it has
        generated a suitable target representation.

        Parameters
        ----------
        new_target : :map:`PointCloud`
            The new target that should be set.
        """
        self._verify_target(new_target)
        self._target_setter(new_target)

    def _verify_target(self, new_target):
        r"""Performs sanity checks to ensure that the new target is valid.

        This includes checking the dimensionality matches and the number of
        points matches the current target's values.

        Parameters
        ----------
        new_target : :map:`PointCloud`
            The target that needs to be verified.

        Raises
        ------
        ValueError
            If the ``new_target`` has differing ``n_points`` or ``n_dims`` to
            ``self``.
        """
        # If the target is None (i.e. on construction) then dodge the
        # verification
        if self.target is None:
            return
        if new_target.n_dims != self.target.n_dims:
            raise ValueError(
                "The current target is {}D, the new target is {}D - new "
                "target has to have the same dimensionality as the "
                "old".format(self.target.n_dims, new_target.n_dims))
        elif new_target.n_points != self.target.n_points:
            raise ValueError(
                "The current target has {} points, the new target has {} "
                "- new target has to have the same number of points as the"
                " old".format(self.target.n_points, new_target.n_points))

    @abc.abstractmethod
    def _target_setter(self, new_target):
        r"""Sets the target to the new value.

        Does no synchronization. Note that it is advisable that
        :meth:`_target_setter_with_verification` is called from
        subclasses instead of this.

        Parameters
        ----------
        new_target : :map:`PointCloud`
            The new target that will be set.
        """

    def _sync_target_from_state(self):
        new_target = self._new_target_from_state()
        self._target_setter_with_verification(new_target)

    @abc.abstractmethod
    def _new_target_from_state(self):
        r"""Generate a new target that is correct after changes to the object.

        Returns
        -------
        object : ``type(self)``
        """
        pass

    @abc.abstractmethod
    def _sync_state_from_target(self):
        r"""Synchronizes the object state to be correct after changes to the
        target.

        Called automatically from the target setter. This is called after the
        target is updated - only handle synchronization here.
        """
        pass


def menpo_src_dir_path():
    r"""The path to the top of the menpo Python package.

    Useful for locating where the data folder is stored.

    Returns
    -------
    path : ``pathlib.Path``
        The full path to the top of the Menpo package
    """
    from pathlib import Path  # to avoid cluttering the menpo.base namespace
    return Path(os.path.abspath(__file__)).parent


class MenpoDeprecationWarning(Warning):
    r"""
    A warning that functionality in Menpo will be deprecated in a future major
    release.
    """
    pass



import numpy as np
from mock import patch, PropertyMock
from nose.tools import raises

import menpo.io as mio
from menpo.image import Image


test_lg = mio.import_landmark_file(mio.data_path_to('breakingbad.pts'))
test_img = Image(np.random.random([100, 100]))
fake_path = '/tmp/test.fake'


@patch('menpo.io.output.base.landmark_types')
@patch('menpo.io.output.base.Path.exists')
@patch('menpo.io.output.base.Path.open')
def test_export_filepath_overwrite_exists(mock_open, exists, landmark_types):
    exists.return_value = True
    mio.export_landmark_file(test_lg, fake_path, overwrite=True)
    mock_open.assert_called_once_with('wb')
    landmark_types.__getitem__.assert_called_once_with('.fake')
    export_function = landmark_types.__getitem__.return_value
    export_function.assert_called_once()


@patch('menpo.io.output.base.landmark_types')
@patch('menpo.io.output.base.Path.exists')
@patch('menpo.io.output.base.Path.open')
def test_export_filepath_no_overwrite(mock_open, exists, landmark_types):
    exists.return_value = False
    mio.export_landmark_file(test_lg, fake_path)
    mock_open.assert_called_once_with('wb')
    landmark_types.__getitem__.assert_called_once_with('.fake')
    export_function = landmark_types.__getitem__.return_value
    export_function.assert_called_once()


@raises(ValueError)
@patch('menpo.io.output.base.landmark_types')
@patch('menpo.io.output.base.Path.exists')
@patch('menpo.io.output.base.Path.open')
def test_export_filepath_wrong_extension(mock_open, exists, landmark_types):
    exists.return_value = False
    mio.export_landmark_file(test_lg, fake_path, extension='pts')
    mock_open.assert_called_once_with('wb')


@patch('menpo.io.output.base.landmark_types')
@patch('menpo.io.output.base.Path.exists')
@patch('menpo.io.output.base.Path.open')
def test_export_filepath_explicit_ext_no_dot(mock_open, exists, landmark_types):
    exists.return_value = False
    mio.export_landmark_file(test_lg, fake_path, extension='fake')
    mock_open.assert_called_once_with('wb')
    landmark_types.__getitem__.assert_called_once_with('.fake')
    export_function = landmark_types.__getitem__.return_value
    export_function.assert_called_once()


@patch('menpo.io.output.base.landmark_types')
@patch('menpo.io.output.base.Path.exists')
@patch('menpo.io.output.base.Path.open')
def test_export_filepath_explicit_ext_dot(mock_open, exists, landmark_types):
    exists.return_value = False
    mio.export_landmark_file(test_lg, fake_path, extension='.fake')
    mock_open.assert_called_once_with('wb')
    landmark_types.__getitem__.assert_called_once_with('.fake')
    export_function = landmark_types.__getitem__.return_value
    export_function.assert_called_once()


@raises(ValueError)
@patch('menpo.io.output.base.Path.exists')
def test_export_filepath_no_overwrite_exists(exists):
    exists.return_value = True
    mio.export_landmark_file(test_lg, fake_path)


@raises(ValueError)
@patch('menpo.io.output.base.landmark_types')
@patch('menpo.io.output.base.Path.exists')
def test_export_unsupported_extension(exists, landmark_types):
    exists.return_value = False
    landmark_types.__getitem__.side_effect = KeyError
    mio.export_landmark_file(test_lg, fake_path)


@raises(ValueError)
@patch('{}.open'.format(__name__), create=True)
def test_export_file_handle_file_extension_None(mock_open):
    with open(fake_path) as f:
        mio.export_landmark_file(test_lg, f)


@raises(ValueError)
@patch('menpo.io.output.base.Path.exists')
@patch('{}.open'.format(__name__), create=True)
def test_export_file_handle_file_extension_not_match_no_dot(mock_open, exists):
    exists.return_value = False
    with open(fake_path) as f:
        type(f).name = PropertyMock(return_value=fake_path)
        mio.export_landmark_file(test_lg, f, extension='pts')
    mock_open.name.assert_called_once()


@raises(ValueError)
@patch('menpo.io.output.base.Path.exists')
@patch('{}.open'.format(__name__), create=True)
def test_export_file_handle_file_extension_not_match_dot(mock_open, exists):
    exists.return_value = False
    with open(fake_path) as f:
        type(f).name = PropertyMock(return_value=fake_path)
        mio.export_landmark_file(test_lg, f, extension='.pts')
    mock_open.name.assert_called_once()


@raises(ValueError)
@patch('menpo.io.output.base.Path.exists')
@patch('{}.open'.format(__name__), create=True)
def test_export_file_handle_file_exists(mock_open, exists):
    exists.return_value = True
    with open(fake_path) as f:
        type(f).name = PropertyMock(return_value=fake_path)
        mio.export_landmark_file(test_lg, f, extension='fake')


@patch('menpo.io.output.base.landmark_types')
@patch('menpo.io.output.base.Path.exists')
@patch('{}.open'.format(__name__), create=True)
def test_export_file_handle_file_exists_overwrite(mock_open, exists,
                                                  landmark_types):
    exists.return_value = True
    with open(fake_path) as f:
        type(f).name = PropertyMock(return_value=fake_path)
        mio.export_landmark_file(test_lg, f, overwrite=True, extension='fake')
    landmark_types.__getitem__.assert_called_once_with('.fake')
    export_function = landmark_types.__getitem__.return_value
    export_function.assert_called_once()


@patch('menpo.io.output.base.landmark_types')
@patch('menpo.io.output.base.Path.exists')
@patch('{}.open'.format(__name__), create=True)
def test_export_file_handle_file_non_file_buffer(mock_open, exists,
                                                 landmark_types):
    exists.return_value = False
    with open(fake_path) as f:
        del f.name  # Equivalent to raising an AttributeError side effect
        mio.export_landmark_file(test_lg, f, extension='fake')
    landmark_types.__getitem__.assert_called_once_with('.fake')
    export_function = landmark_types.__getitem__.return_value
    export_function.assert_called_once()


@patch('menpo.io.output.landmark.json.dump')
@patch('menpo.io.output.base.Path.exists')
@patch('{}.open'.format(__name__), create=True)
def test_export_landmark_ljson(mock_open, exists, json_dump):
    exists.return_value = False
    fake_path = '/fake/fake.ljson'
    with open(fake_path) as f:
        type(f).name = PropertyMock(return_value=fake_path)
        mio.export_landmark_file(test_lg, f, extension='ljson')
    json_dump.assert_called_once()


@patch('menpo.io.output.landmark.np.savetxt')
@patch('menpo.io.output.base.Path.exists')
@patch('{}.open'.format(__name__), create=True)
def test_export_landmark_pts(mock_open, exists, save_txt):
    exists.return_value = False
    fake_path = '/fake/fake.pts'
    with open(fake_path) as f:
        type(f).name = PropertyMock(return_value=fake_path)
        mio.export_landmark_file(test_lg, f, extension='pts')
    save_txt.assert_called_once()


@patch('menpo.image.base.PILImage')
@patch('menpo.io.output.base.Path.exists')
@patch('{}.open'.format(__name__), create=True)
def test_export_image_jpg(mock_open, exists, PILImage):
    exists.return_value = False
    fake_path = '/fake/fake.jpg'
    with open(fake_path) as f:
        type(f).name = PropertyMock(return_value=fake_path)
        mio.export_image(test_img, f, extension='jpg')
    PILImage.save.assert_called_once()


@patch('menpo.io.output.pickle.pickle.dump')
@patch('menpo.io.output.base.Path.exists')
@patch('{}.open'.format(__name__), create=True)
def test_export_pickle(mock_open, exists, pickle_dump):
    exists.return_value = False
    fake_path = '/fake/fake.pkl'
    with open(fake_path) as f:
        type(f).name = PropertyMock(return_value=fake_path)
        mio.export_pickle(test_lg, f)
    pickle_dump.assert_called_once()


@patch('menpo.io.output.pickle.pickle.dump')
@patch('menpo.io.output.base.Path.exists')
@patch('__builtin__.open')
def test_export_pickle_with_path_uses_open(mock_open, exists, pickle_dump):
    exists.return_value = False
    fake_path = '/fake/fake.pkl.gz'
    mio.export_pickle(test_lg, fake_path)
    pickle_dump.assert_called_once()
    mock_open.assert_called_once_with(fake_path, 'wb')



"""
This module is only designed for use inside of our testing. It isn't used or
exposed anywhere except in our tests. It is useful, because it contains general
methods that are applicable across many of our tests.
"""


def is_same_array(a, b):
    """
    Check if `a` and `b` represent the same piece of memory.

    Parameters
    ----------
    a : ndarray
        First array to compare
    b : ndarray
        Second array to compare

    Returns
    -------
    is_same : bool
        Will be `True` if the two arrays represent the same piece of memory.
    """
    if not a.flags['OWNDATA'] and not b.flags['OWNDATA']:
        return a.base is b.base
    if not a.flags['OWNDATA'] and b.flags['OWNDATA']:
        return a.base is b
    if not b.flags['OWNDATA'] and a.flags['OWNDATA']:
        return b.base is a

    # Fallthough, they are either the same array or they aren't!
    return a is b



# This file was generated by 'versioneer.py' (0.10) from
# revision-control system data, or from the parent directory name of an
# unpacked source archive. Distribution tarballs contain a pre-generated copy
# of this file.

version_version = '0.4.0'
version_full = '04277b9c50dafb8d51773d84b40b9582c166e31d'
def get_versions(default={}, verbose=False):
    return {'version': version_version, 'full': version_full}




import abc
from collections import OrderedDict
import json
import warnings

import numpy as np

from menpo.landmark.base import LandmarkGroup
from menpo.shape import PointCloud, PointUndirectedGraph
from menpo.transform import Scale
from .base import Importer


class LandmarkImporter(Importer):
    """
    Abstract base class for importing landmarks.

    Parameters
    ----------
    filepath : string
        Absolute filepath of the landmarks.
    """

    __metaclass__ = abc.ABCMeta

    def __init__(self, filepath):
        super(LandmarkImporter, self).__init__(filepath)
        self.pointcloud = None
        self.labels_to_masks = None

    def build(self, asset=None):
        """
        Overrides the :meth:`build <menpo.io.base.Importer.build>` method.

        Parse the landmark format and return the label and landmark dictionary.

        Parameters
        ----------
        asset : object, optional
            The asset that the landmarks are being built for. Can be used to
            adjust landmarks as necessary (e.g. rescaling image landmarks
            from 0-1 to image.shape)

        Returns
        -------
        landmark_group : string
            The landmark group parsed from the file.
            Every point will be labelled.
        """
        self._parse_format(asset=asset)
        return LandmarkGroup(self.pointcloud,
                             self.labels_to_masks)

    @abc.abstractmethod
    def _parse_format(self, asset=None):
        r"""
        Read the landmarks file from disk, parse it in to semantic labels and
        :class:`menpo.shape.base.PointCloud`.

        Set the `self.label` and `self.pointcloud` attributes.
        """
        pass


class ASFImporter(LandmarkImporter):
    r"""
    Abstract base class for an importer for the ASF file format.
    Currently **does not support the connectivity specified in the format**.

    Implementations of this class should override the :meth:`_build_points`
    which determines the ordering of axes. For example, for images, the
    `x` and `y` axes are flipped such that the first axis is `y` (height
    in the image domain).

    Landmark set label: ASF

    Landmark labels:

    +---------+
    | label   |
    +=========+
    | all     |
    +---------+

    Parameters
    ----------
    filepath : string
        Absolute filepath to landmark file.

    References
    ----------
    .. [1] http://www2.imm.dtu.dk/~aam/datasets/datasets.html
    """
    def __init__(self, filepath):
        super(ASFImporter, self).__init__(filepath)

    @abc.abstractmethod
    def _build_points(self, xs, ys):
        r"""
        Determines the ordering of points within the landmarks. For meshes
        `x` is the first axis, where as for images `y` is the first axis.
        """
        pass

    def _parse_format(self, asset=None):
        with open(self.filepath, 'r') as f:
            landmarks = f.read()

        # Remove comments and blank lines
        landmarks = [l for l in landmarks.splitlines()
                     if (l.rstrip() and not '#' in l)]

        # Pop the front of the list for the number of landmarks
        count = int(landmarks.pop(0))
        # Pop the last element of the list for the image_name
        image_name = landmarks.pop()

        xs = np.empty([count, 1])
        ys = np.empty([count, 1])
        connectivity = np.empty([count, 2], dtype=np.int)
        for i in xrange(count):
            # Though unpacked, they are still all strings
            # Only unpack the first 7
            (path_num, path_type, xpos, ypos,
             point_num, connects_from, connects_to) = landmarks[i].split()[:7]
            xs[i, ...] = float(xpos)
            ys[i, ...] = float(ypos)
            connectivity[i, ...] = [int(connects_from), int(connects_to)]

        points = self._build_points(xs, ys)
        if asset is not None:
            # we've been given an asset. As ASF files are normalized,
            # fix that here
            points = Scale(np.array(asset.shape)).apply(points)

        # TODO: Use connectivity and create a graph type instead of PointCloud
        # edges = scaled_points[connectivity]

        self.pointcloud = PointCloud(points)
        self.labels_to_masks = OrderedDict(
            [('all', np.ones(points.shape[0], dtype=np.bool))])


class PTSImporter(LandmarkImporter):
    r"""
    Importer for the PTS file format. Assumes version 1 of the format.

    Implementations of this class should override the :meth:`_build_points`
    which determines the ordering of axes. For example, for images, the
    `x` and `y` axes are flipped such that the first axis is `y` (height
    in the image domain).

    Note that PTS has a very loose format definition. Here we make the
    assumption (as is common) that PTS landmarks are 1-based. That is,
    landmarks on a 480x480 image are in the range [1-480]. As Menpo is
    consistently 0-based, we *subtract 1* off each landmark value
    automatically.

    If you want to use PTS landmarks that are 0-based, you will have to
    manually add one back on to landmarks post importing.

    Landmark set label: PTS

    Landmark labels:

    +---------+
    | label   |
    +=========+
    | all     |
    +---------+
    """
    def __init__(self, filepath):
        super(PTSImporter, self).__init__(filepath)

    @abc.abstractmethod
    def _build_points(self, xs, ys):
        r"""
        Determines the ordering of points within the landmarks. For meshes
        `x` is the first axis, where as for images `y` is the first axis.
        """
        pass

    def _parse_format(self, asset=None):
        f = open(self.filepath, 'r')
        for line in f:
            if line.split()[0] == '{':
                break
        xs = []
        ys = []
        for line in f:
            if line.split()[0] != '}':
                xpos, ypos = line.split()[0:2]
                xs.append(xpos)
                ys.append(ypos)
        xs = np.array(xs, dtype=np.float).reshape((-1, 1))
        ys = np.array(ys, dtype=np.float).reshape((-1, 1))
        # PTS landmarks are 1-based, need to convert to 0-based (subtract 1)
        points = self._build_points(xs - 1, ys - 1)

        self.pointcloud = PointCloud(points)
        self.labels_to_masks = OrderedDict(
            [('all', np.ones(points.shape[0], dtype=np.bool))])


class LM2Importer(LandmarkImporter):
    r"""
    Importer for the LM2 file format from the bosphorus dataset. This is a 2D
    landmark type and so it is assumed it only applies to images.

    Landmark set label: LM2

    Landmark labels:

    +------------------------+
    | label                  |
    +========================+
    | outer_left_eyebrow     |
    | middle_left_eyebrow    |
    | inner_left_eyebrow     |
    | inner_right_eyebrow    |
    | middle_right_eyebrow   |
    | outer_right_eyebrow    |
    | outer_left_eye_corner  |
    | inner_left_eye_corner  |
    | inner_right_eye_corner |
    | outer_right_eye_corner |
    | nose_saddle_left       |
    | nose_saddle_right      |
    | left_nose_peak         |
    | nose_tip               |
    | right_nose_peak        |
    | left_mouth_corner      |
    | upper_lip_outer_middle |
    | right_mouth_corner     |
    | upper_lip_inner_middle |
    | lower_lip_inner_middle |
    | lower_lip_outer_middle |
    | chin_middle            |
    +------------------------+
    """
    def __init__(self, filepath):
        super(LM2Importer, self).__init__(filepath)

    def _parse_format(self, asset=None):
        with open(self.filepath, 'r') as f:
            landmarks = f.read()

        # Remove comments and blank lines
        landmark_text = [l for l in landmarks.splitlines()
                         if (l.rstrip() and '#' not in l)]

        # First line says how many landmarks there are: 22 Landmarks
        # So pop it off the front
        num_points = int(landmark_text.pop(0).split()[0])
        labels = []

        # The next set of lines defines the labels
        labels_str = landmark_text.pop(0)
        if not labels_str == 'Labels:':
            raise ImportError("LM2 landmarks are incorrectly formatted. "
                              "Expected a list of labels beginning with "
                              "'Labels:' but found '{0}'".format(labels_str))
        for i in xrange(num_points):
            # Lowercase, remove spaces and replace with underscores
            l = landmark_text.pop(0)
            l = '_'.join(l.lower().split())
            labels.append(l)

        # The next set of lines defines the coordinates
        coords_str = landmark_text.pop(0)
        if not coords_str == '2D Image coordinates:':
            raise ImportError("LM2 landmarks are incorrectly formatted. "
                              "Expected a list of coordinates beginning with "
                              "'2D Image coordinates:' "
                              "but found '{0}'".format(coords_str))
        xs = []
        ys = []
        for i in xrange(num_points):
            p = landmark_text.pop(0).split()
            xs.append(float(p[0]))
            ys.append(float(p[1]))

        xs = np.array(xs, dtype=np.float).reshape((-1, 1))
        ys = np.array(ys, dtype=np.float).reshape((-1, 1))

        # Flip the x and y
        self.pointcloud = PointCloud(np.hstack([ys, xs]))
        # Create the mask whereby there is one landmark per label
        # (identity matrix)
        masks = np.eye(num_points).astype(np.bool)
        masks = np.vsplit(masks, num_points)
        masks = [np.squeeze(m) for m in masks]
        self.labels_to_masks = OrderedDict(zip(labels, masks))


def _parse_ljson_v1(lms_dict):
    from menpo.base import MenpoDeprecationWarning
    warnings.warn('LJSON v1 is deprecated. export_landmark_file{s}() will '
                  'only save out LJSON v2 files. Please convert all LJSON '
                  'files to v2 by importing into Menpo and re-exporting to '
                  'overwrite the files.', MenpoDeprecationWarning)
    all_points = []
    labels = []  # label per group
    labels_slices = []  # slices into the full pointcloud per label
    offset = 0
    connectivity = []
    for group in lms_dict['groups']:
        lms = group['landmarks']
        labels.append(group['label'])
        labels_slices.append(slice(offset, len(lms) + offset))
        # Create the connectivity if it exists
        conn = group.get('connectivity', [])
        if conn:
            # Offset relative connectivity according to the current index
            conn = offset + np.asarray(conn)
            connectivity.append(conn)
        for p in lms:
            all_points.append(p['point'])
        offset += len(lms)

    # Don't create a PointUndirectedGraph with no connectivity
    points = np.array(all_points)
    if len(connectivity) == 0:
        pcloud = PointCloud(points)
    else:
        pcloud = PointUndirectedGraph(points, np.vstack(connectivity))
    labels_to_masks = OrderedDict()
    # go through each label and build the appropriate boolean array
    for label, l_slice in zip(labels, labels_slices):
        mask = np.zeros(pcloud.n_points, dtype=np.bool)
        mask[l_slice] = True
        labels_to_masks[label] = mask
    return pcloud, labels_to_masks


def _parse_ljson_v2(lms_dict):
    labels_to_mask = OrderedDict()  # masks into the full pointcloud per label

    points = np.array(lms_dict['landmarks']['points'])
    connectivity = lms_dict['landmarks'].get('connectivity')

    # Don't create a PointUndirectedGraph with no connectivity
    if connectivity is None:
        pcloud = PointCloud(points)
    else:
        pcloud = PointUndirectedGraph(points, np.vstack(connectivity))

    for label in lms_dict['labels']:
        mask = np.zeros(pcloud.n_points, dtype=np.bool)
        mask[label['mask']] = True
        labels_to_mask[label['label']] = mask

    return pcloud, labels_to_mask


_ljson_parser_for_version = {
    1: _parse_ljson_v1,
    2: _parse_ljson_v2
}


class LJSONImporter(LandmarkImporter):
    r"""
    Importer for the Menpo JSON format. This is an n-dimensional
    landmark type for both images and meshes that encodes semantic labels in
    the format.

    Landmark set label: JSON

    Landmark labels: decided by file

    """
    def _parse_format(self, asset=None):
        with open(self.filepath, 'rb') as f:
            # lms_dict is now a dict rep of the JSON
            lms_dict = json.load(f, object_pairs_hook=OrderedDict)
        v = lms_dict.get('version')
        parser = _ljson_parser_for_version.get(v)
        if parser is None:
            raise ValueError("{} has unknown version {} must be "
                             "1, or 2".format(self.filepath, v))
        else:
            self.pointcloud, self.labels_to_masks = parser(lms_dict)
